import pandas as pd
import numpy as np
from celldetective.utils import derivative, extract_identity_col
import os
import subprocess
abs_path = os.sep.join([os.path.split(os.path.dirname(os.path.realpath(__file__)))[0], 'celldetective'])

def measure_pairs(pos, neighborhood_protocol):
	
	reference_population = neighborhood_protocol['reference']
	neighbor_population = neighborhood_protocol['neighbor']
	neighborhood_type = neighborhood_protocol['type']
	neighborhood_distance = neighborhood_protocol['distance']
	neighborhood_description = neighborhood_protocol['description']

	relative_measurements = []

	tab_ref = pos + os.sep.join(['output', 'tables', f'trajectories_{reference_population}.pkl'])
	if os.path.exists(tab_ref):
		df_reference = np.load(tab_ref, allow_pickle=True)
	else:
		df_reference = None

	if os.path.exists(tab_ref.replace(reference_population, neighbor_population)):
		df_neighbor = np.load(tab_ref.replace(reference_population, neighbor_population), allow_pickle=True)
	else:
		if os.path.exists(tab_ref.replace(reference_population, neighbor_population).replace('.pkl','.csv')):
			df_neighbor = pd.read_csv(tab_ref.replace(reference_population, neighbor_population).replace('.pkl','.csv'))
		else:
			df_neighbor = None

	if df_reference is None:
		return None

	assert str(neighborhood_description) in list(df_reference.columns)
	neighborhood = df_reference.loc[:,f'{neighborhood_description}'].to_numpy()

	ref_id_col = extract_identity_col(df_reference)
	ref_tracked = False
	if ref_id_col is None:
		return None
	elif ref_id_col=='TRACK_ID':
		ref_tracked = True
	neigh_id_col = extract_identity_col(df_neighbor)
	neigh_tracked = False
	if neigh_id_col is None:
		return None
	elif neigh_id_col=='TRACK_ID':
		neigh_tracked = True
	
	centre_of_mass_columns = [(c,c.replace('POSITION_X','POSITION_Y')) for c in list(df_neighbor.columns) if c.endswith('centre_of_mass_POSITION_X')]
	centre_of_mass_labels = [c.replace('_centre_of_mass_POSITION_X','') for c in list(df_neighbor.columns) if c.endswith('centre_of_mass_POSITION_X')]

	for t in np.unique(list(df_reference['FRAME'].unique())+list(df_neighbor['FRAME'])):
		
		group_reference = df_reference.loc[df_reference['FRAME']==t,:]
		group_neighbors = df_neighbor.loc[df_neighbor['FRAME']==t, :]
		
		for tid, group in group_reference.groupby(ref_id_col):

			neighborhood = group.loc[: , f'{neighborhood_description}'].to_numpy()[0]
			coords_reference = group[['POSITION_X', 'POSITION_Y']].to_numpy()[0]

			neighbors = []
			if isinstance(neighborhood, float) or neighborhood!=neighborhood:
				pass
			else:
				for neigh in neighborhood:
					neighbors.append(neigh['id'])

			unique_neigh = list(np.unique(neighbors))
			print(f'{unique_neigh=}')

			neighbor_properties = group_neighbors.loc[group_neighbors[neigh_id_col].isin(unique_neigh)]
			
			for nc, group_neigh in neighbor_properties.groupby(neigh_id_col):
				
				neighbor_vector = np.zeros((2))
				neighbor_vector[:] = np.nan
				mass_displacement_vector = np.zeros((len(centre_of_mass_columns), 2))

				coords_centre_of_mass = []
				for col in centre_of_mass_columns:
					coords_centre_of_mass.append(group_neigh[[col[0],col[1]]].to_numpy()[0])

				dot_product_vector = np.zeros((len(centre_of_mass_columns)))
				dot_product_vector[:] = np.nan

				cosine_dot_vector = np.zeros((len(centre_of_mass_columns)))
				cosine_dot_vector[:] = np.nan

				coords_neighbor = group_neigh[['POSITION_X', 'POSITION_Y']].to_numpy()[0]
				neighbor_vector[0] = coords_neighbor[0] - coords_reference[0]
				neighbor_vector[1] = coords_neighbor[1] - coords_reference[1]

				if neighbor_vector[0]==neighbor_vector[0] and neighbor_vector[1]==neighbor_vector[1]:
					angle = np.arctan2(neighbor_vector[1], neighbor_vector[0])
					relative_distance = np.sqrt(neighbor_vector[0]**2 + neighbor_vector[1]**2)
					
					for z,cols in enumerate(centre_of_mass_columns):

							mass_displacement_vector[z,0] = coords_centre_of_mass[z][0] - coords_neighbor[0]
							mass_displacement_vector[z,1] = coords_centre_of_mass[z][1] - coords_neighbor[1]

							dot_product_vector[z] = np.dot(mass_displacement_vector[z], -neighbor_vector)
							cosine_dot_vector[z] = np.dot(mass_displacement_vector[z], -neighbor_vector) / (np.linalg.norm(mass_displacement_vector[z])*np.linalg.norm(-neighbor_vector))

					relative_measurements.append(
							{'REFERENCE_ID': tid, 'NEIGHBOR_ID': nc,
							'reference_population': reference_population,
							'neighbor_population': neighbor_population,
							'FRAME': t, 'distance': relative_distance,
							'angle': angle * 180 / np.pi,
							f'status_{neighborhood_description}': 1,
							f'class_{neighborhood_description}': 0,
							'reference_tracked': ref_tracked, 'neighbors_tracked': neigh_tracked,
							 })
					for z,lbl in enumerate(centre_of_mass_labels):
						relative_measurements[-1].update({lbl+'_centre_of_mass_dot_product': dot_product_vector[z], lbl+'_centre_of_mass_dot_cosine': cosine_dot_vector[z]})
	
	df_pairs = pd.DataFrame(relative_measurements)
	
	return df_pairs







def measure_pair_signals_at_position(pos, neighborhood_protocol, velocity_kwargs={'window': 3, 'mode': 'bi'}):
	"""
	pos: position to process
	target_classes [list]: target classes to keep
	neigh_dist: neighborhood cut distance
	theta_dist: distance to edge threshold
	velocity_kwargs: params for derivative of relative position
	neighborhood_kwargs: params for neigh
	"""

	reference_population = neighborhood_protocol['reference']
	neighbor_population = neighborhood_protocol['neighbor']
	neighborhood_type = neighborhood_protocol['type']
	neighborhood_distance = neighborhood_protocol['distance']
	neighborhood_description = neighborhood_protocol['description']

	relative_measurements = []

	tab_ref = pos + os.sep.join(['output', 'tables', f'trajectories_{reference_population}.pkl'])
	if os.path.exists(tab_ref):
		df_reference = np.load(tab_ref, allow_pickle=True)
	else:
		df_reference = None

	if os.path.exists(tab_ref.replace(reference_population, neighbor_population)):
		df_neighbor = np.load(tab_ref.replace(reference_population, neighbor_population), allow_pickle=True)
	else:
		if os.path.exists(tab_ref.replace(reference_population, neighbor_population).replace('.pkl','.csv')):
			df_neighbor = pd.read_csv(tab_ref.replace(reference_population, neighbor_population).replace('.pkl','.csv'))
		else:
			df_neighbor = None

	if df_reference is None:
		return None

	assert str(neighborhood_description) in list(df_reference.columns)
	neighborhood = df_reference.loc[:,f'{neighborhood_description}'].to_numpy()

	ref_id_col = extract_identity_col(df_reference)
	if ref_id_col is not None:
		df_reference = df_reference.sort_values(by=[ref_id_col, 'FRAME'])

	ref_tracked = False
	if ref_id_col=='TRACK_ID':
		compute_velocity = True
		ref_tracked = True
	elif ref_id_col=='ID':
		df_pairs = measure_pairs(pos, neighborhood_protocol)
		return df_pairs
	else:
		print('ID or TRACK ID column could not be found in neighbor table. Abort.')
		return None

	print(f'Measuring pair signals...')

	neigh_id_col = extract_identity_col(df_neighbor)
	neigh_tracked = False
	if neigh_id_col=='TRACK_ID':
		compute_velocity = True
		neigh_tracked = True
	elif neigh_id_col=='ID':
		df_pairs = measure_pairs(pos, neighborhood_protocol)
		return df_pairs
	else:
		print('ID or TRACK ID column could not be found in neighbor table. Abort.')
		return None

	try:
		for tid, group in df_reference.groupby(ref_id_col):

			neighbor_dicts = group.loc[: , f'{neighborhood_description}'].values
			timeline_reference = group['FRAME'].to_numpy()
			coords_reference = group[['POSITION_X', 'POSITION_Y']].to_numpy()

			neighbor_ids = []
			neighbor_ids_per_t = []

			time_of_first_entrance_in_neighborhood = {}
			t_departure={}

			for t in range(len(timeline_reference)):

				neighbors_at_t = neighbor_dicts[t]
				neighs_t = []
				if isinstance(neighbors_at_t, float) or neighbors_at_t!=neighbors_at_t:
					pass
				else:
					for neigh in neighbors_at_t:
						if neigh['id'] not in neighbor_ids:
							time_of_first_entrance_in_neighborhood[neigh['id']]=t
						neighbor_ids.append(neigh['id'])
						neighs_t.append(neigh['id'])
				neighbor_ids_per_t.append(neighs_t)

			#print(neighbor_ids_per_t)
			unique_neigh = list(np.unique(neighbor_ids))
			print(f'Reference cell {tid}: found {len(unique_neigh)} neighbour cells: {unique_neigh}...')

			neighbor_properties = df_neighbor.loc[df_neighbor[neigh_id_col].isin(unique_neigh)]

			for nc, group_neigh in neighbor_properties.groupby(neigh_id_col):
				
				coords_neighbor = group_neigh[['POSITION_X', 'POSITION_Y']].to_numpy()
				timeline_neighbor = group_neigh['FRAME'].to_numpy()

				# # Perform timeline matching to have same start-end points and no gaps
				full_timeline, _, _ = timeline_matching(timeline_reference, timeline_neighbor)

				neighbor_vector = np.zeros((len(full_timeline), 2))
				neighbor_vector[:,:] = np.nan

				centre_of_mass_columns = [(c,c.replace('POSITION_X','POSITION_Y')) for c in list(neighbor_properties.columns) if c.endswith('centre_of_mass_POSITION_X')]
				centre_of_mass_labels = [c.replace('_centre_of_mass_POSITION_X','') for c in list(neighbor_properties.columns) if c.endswith('centre_of_mass_POSITION_X')]

				mass_displacement_vector = np.zeros((len(centre_of_mass_columns), len(full_timeline), 2))
				mass_displacement_vector[:,:,:] = np.nan

				dot_product_vector = np.zeros((len(centre_of_mass_columns), len(full_timeline)))
				dot_product_vector[:,:] = np.nan

				cosine_dot_vector = np.zeros((len(centre_of_mass_columns), len(full_timeline)))
				cosine_dot_vector[:,:] = np.nan

				coords_centre_of_mass = []
				for col in centre_of_mass_columns:
					coords_centre_of_mass.append(group_neigh[[col[0],col[1]]].to_numpy())

				# Relative distance
				for t in range(len(full_timeline)):

					if t in timeline_reference and t in timeline_neighbor: # meaning position exists on both sides

						idx_reference = list(timeline_reference).index(t) #index_reference[list(full_timeline).index(t)]
						idx_neighbor = list(timeline_neighbor).index(t) #index_neighbor[list(full_timeline).index(t)]

						neighbor_vector[t, 0] = coords_neighbor[idx_neighbor, 0] - coords_reference[idx_reference, 0]
						neighbor_vector[t, 1] = coords_neighbor[idx_neighbor, 1] - coords_reference[idx_reference, 1]

						for z,cols in enumerate(centre_of_mass_columns):

							mass_displacement_vector[z,t,0] = coords_centre_of_mass[z][idx_neighbor, 0] - coords_neighbor[idx_neighbor, 0]
							mass_displacement_vector[z,t,1] = coords_centre_of_mass[z][idx_neighbor, 1] - coords_neighbor[idx_neighbor, 1]

							dot_product_vector[z,t] = np.dot(mass_displacement_vector[z,t], -neighbor_vector[t])
							cosine_dot_vector[z,t] = np.dot(mass_displacement_vector[z,t], -neighbor_vector[t]) / (np.linalg.norm(mass_displacement_vector[z,t])*np.linalg.norm(-neighbor_vector[t]))
							if tid==44.0 and nc==173.0:
								print(f'{centre_of_mass_columns[z]=} {mass_displacement_vector[z,t]=} {-neighbor_vector[t]=} {dot_product_vector[z,t]=} {cosine_dot_vector[z,t]=}')
							

				angle = np.zeros(len(full_timeline))
				angle[:] = np.nan

				exclude = neighbor_vector[:,1]!=neighbor_vector[:,1]
				angle[~exclude] = np.arctan2(neighbor_vector[:, 1][~exclude], neighbor_vector[:, 0][~exclude])
				#print(f'Angle before unwrap: {angle}')
				angle[~exclude] = np.unwrap(angle[~exclude])
				#print(f'Angle after unwrap: {angle}')
				relative_distance = np.sqrt(neighbor_vector[:,0]**2 + neighbor_vector[:, 1]**2)
				#print(f'Timeline: {full_timeline}; Distance: {relative_distance}')

				if compute_velocity:
					rel_velocity = derivative(relative_distance, full_timeline, **velocity_kwargs)
					rel_velocity_long_timescale = derivative(relative_distance, full_timeline, window = 7, mode='bi')
					#rel_velocity = np.insert(rel_velocity, 0, np.nan)[:-1]
					
					angular_velocity = np.zeros(len(full_timeline))
					angular_velocity[:] = np.nan
					angular_velocity_long_timescale = np.zeros(len(full_timeline))
					angular_velocity_long_timescale[:] = np.nan

					angular_velocity[~exclude] = derivative(angle[~exclude], full_timeline[~exclude], **velocity_kwargs)
					angular_velocity_long_timescale[~exclude] = derivative(angle[~exclude], full_timeline[~exclude], window = 7, mode='bi')

				# 	angular_velocity = np.zeros(len(full_timeline))
				# 	angular_velocity[:] = np.nan

				# 	for t in range(1, len(relative_angle1)):
				# 		if not np.isnan(relative_angle1[t]) and not np.isnan(relative_angle1[t - 1]):
				# 			delta_angle = relative_angle1[t] - relative_angle1[t - 1]
				# 			delta_time = full_timeline[t] - full_timeline[t - 1]
				# 			if delta_time != 0:
				# 				angular_velocity[t] = delta_angle / delta_time

				duration_in_neigh = list(neighbor_ids).count(nc)
				#print(nc, duration_in_neigh, ' frames')

				cum_sum = 0
				for t in range(len(full_timeline)):

					if t in timeline_reference: # meaning position exists on both sides

						idx_reference = list(timeline_reference).index(t) 

						if nc in neighbor_ids_per_t[idx_reference]:

							cum_sum+=1
							relative_measurements.append(
									{'REFERENCE_ID': tid, 'NEIGHBOR_ID': nc,
									'reference_population': reference_population,
									'neighbor_population': neighbor_population,
									'FRAME': t, 'distance': relative_distance[t],
									'velocity': rel_velocity[t],
									'velocity_smooth': rel_velocity_long_timescale[t], 
									'angle': angle[t] * 180 / np.pi,
									#'angle-neigh-ref': angle[t] * 180 / np.pi, 
									'angular_velocity': angular_velocity[t],
									'angular_velocity_smooth': angular_velocity_long_timescale[t],
									f'status_{neighborhood_description}': 1,
									f'residence_time_in_{neighborhood_description}': cum_sum,
									f'class_{neighborhood_description}': 0,
									f't0_{neighborhood_description}': time_of_first_entrance_in_neighborhood[nc],
									'reference_tracked': ref_tracked, 'neighbors_tracked': neigh_tracked,
									 })
							for z,lbl in enumerate(centre_of_mass_labels):
								relative_measurements[-1].update({lbl+'_centre_of_mass_dot_product': dot_product_vector[z,t], lbl+'_centre_of_mass_dot_cosine': cosine_dot_vector[z,t]})

						else:
							relative_measurements.append(
									{'REFERENCE_ID': tid, 'NEIGHBOR_ID': nc,
									'reference_population': reference_population,
									'neighbor_population': neighbor_population,
									'FRAME': t, 'distance': relative_distance[t],
									'velocity': rel_velocity[t], 
									'velocity_smooth': rel_velocity_long_timescale[t],
									'angle': angle[t] * 180 / np.pi,
									#'angle-neigh-ref': angle[t] * 180 / np.pi, 
									'angular_velocity': angular_velocity[t],
									'angular_velocity_smooth': angular_velocity_long_timescale[t],
									f'status_{neighborhood_description}': 0,
									f'residence_time_in_{neighborhood_description}': cum_sum,
									f'class_{neighborhood_description}': 0,
									f't0_{neighborhood_description}': time_of_first_entrance_in_neighborhood[nc],
									'reference_tracked': ref_tracked, 'neighbors_tracked': neigh_tracked,
									 })
							for z,lbl in enumerate(centre_of_mass_labels):
								relative_measurements[-1].update({lbl+'_centre_of_mass_dot_product': dot_product_vector[z,t], lbl+'_centre_of_mass_dot_cosine': cosine_dot_vector[z,t]})

		df_pairs = pd.DataFrame(relative_measurements)

		return df_pairs

	except KeyError:
		print(f"Neighborhood {description} not found in data frame. Measurements for this neighborhood will not be calculated")


def timeline_matching(timeline1, timeline2):

	"""
	Match two timelines and create a unified timeline with corresponding indices.

	Parameters
	----------
	timeline1 : array-like
		The first timeline to be matched.
	timeline2 : array-like
		The second timeline to be matched.

	Returns
	-------
	tuple
		A tuple containing:
		- full_timeline : numpy.ndarray
			The unified timeline spanning from the minimum to the maximum time point in the input timelines.
		- index1 : list of int
			The indices of `timeline1` in the `full_timeline`.
		- index2 : list of int
			The indices of `timeline2` in the `full_timeline`.

	Examples
	--------
	>>> timeline1 = [1, 2, 5, 6]
	>>> timeline2 = [2, 3, 4, 6]
	>>> full_timeline, index1, index2 = timeline_matching(timeline1, timeline2)
	>>> print(full_timeline)
	[1 2 3 4 5 6]
	>>> print(index1)
	[0, 1, 4, 5]
	>>> print(index2)
	[1, 2, 3, 5]

	Notes
	-----
	- The function combines the two timelines and generates a continuous range from the minimum to the maximum time point.
	- It then finds the indices of the original timelines in this unified timeline.
	- The function assumes that the input timelines consist of integer values.
	"""

	min_t = np.amin(np.concatenate((timeline1, timeline2)))
	max_t = np.amax(np.concatenate((timeline1, timeline2)))
	full_timeline = np.arange(min_t, max_t + 1)

	index1 = [list(np.where(full_timeline == int(t))[0])[0] for t in timeline1]
	index2 = [list(np.where(full_timeline == int(t))[0])[0] for t in timeline2]

	return full_timeline, index1, index2


def rel_measure_at_position(pos):

	pos = pos.replace('\\', '/')
	pos = rf"{pos}"
	assert os.path.exists(pos), f'Position {pos} is not a valid path.'
	if not pos.endswith('/'):
		pos += '/'
	script_path = os.sep.join([abs_path, 'scripts', 'measure_relative.py'])
	cmd = f'python "{script_path}" --pos "{pos}"'
	subprocess.call(cmd, shell=True)


# def mcf7_size_model(x,x0,x2):
# 	return np.piecewise(x, [x<= x0, (x > x0)*(x<=x2), x > x2], [lambda x: 1, lambda x: -1/(x2-x0)*x + (1+x0/(x2-x0)), 0])
# def sigmoid(x,x0,k):
# 	return 1/(1 + np.exp(-(x-x0)/k))
# def velocity_law(x):
# 	return np.piecewise(x, [x<=-10, x > -10],[lambda x: 0., lambda x: (1*x+10)*(1-sigmoid(x, 1,1))/10])


# def probabilities(pairs,radius_critical=80,radius_max=150):
# 	scores = []
# 	pair_dico=[]
# 	print(f'Found {len(pairs)} TC-NK pairs...')
# 	if len(pairs) > 0:
# 		unique_tcs = np.unique(pairs['tc'].to_numpy())
# 		unique_nks = np.unique(pairs['nk'].to_numpy())
# 		matrix = np.zeros((len(unique_tcs), len(unique_nks)))
# 		for index, row in pairs.iterrows():

# 			i = np.where(unique_tcs == row['tc'])[0]
# 			j = np.where(unique_nks == row['nk'])[0]

# 			d_prob = mcf7_size_model(row['drel'], radius_critical, radius_max)
# 			lamp_prob = sigmoid(row['lamp1'], 1.05, 0.01)
# 			synapse_prob = row['syn_class']
# 			velocity_prob = velocity_law(row['vrel'])  # 1-sigmoid(row['vrel'], 1,1)
# 			time_prob = row['t_residence_rel']

# 			hypotheses = [d_prob, velocity_prob, lamp_prob, synapse_prob,
# 						  time_prob]  # lamp_prob d_prob, synapse_prob, velocity_prob, lamp_prob
# 			s = np.sum(hypotheses) / len(hypotheses)

# 			matrix[i, j] = s  # synapse_prob': synapse_prob,
# 			pair_dico.append(
# 				{ 'tc': row['tc'], 'nk': row['nk'], 'synapse_prob': synapse_prob,
# 				 'd_prob': d_prob, 'lamp_prob': lamp_prob, 'velocity_prob': velocity_prob, 'time_prob': time_prob})
# 		pair_dico = pd.DataFrame(pair_dico)

# 		hypotheses = ['velocity_prob', 'd_prob', 'time_prob', 'lamp_prob', 'synapse_prob']

# 		for i in tqdm(range(2000)):
# 			sample = np.array(random.choices(np.linspace(0, 1, 100), k=len(hypotheses)))
# 			weights = sample / np.sum(sample)

# 			score_i = {}
# 			for k, hyp in enumerate(hypotheses):
# 				score_i.update({'w_' + hyp: weights[k]})
# 			probs=[]
# 			for cells, group in pair_dico.groupby(['tc']):

# 				group['total_prob'] = 0
# 				for hyp in hypotheses:
# 					group['total_prob'] += group[hyp] * score_i['w_' + hyp]
# 					probs.append(group)
# 	return probs

def update_effector_table(df_relative, df_effector):
	df_effector['group_neighborhood']=1
	effectors = np.unique(df_relative['EFFECTOR_ID'].to_numpy())
	for effector in effectors:
		try:
			# Set group_neighborhood to 0 where TRACK_ID matches effector
			df_effector.loc[df_effector['TRACK_ID'] == effector, 'group_neighborhood'] = 0
		except:
			df_effector.loc[df_effector['ID'] == effector, 'group_neighborhood'] = 0
	return df_effector

def extract_neighborhoods_from_pickles(pos):

	"""
	Extract neighborhood protocols from pickle files located at a given position.

	Parameters
	----------
	pos : str
		The base directory path where the pickle files are located.

	Returns
	-------
	list of dict
		A list of dictionaries, each containing a neighborhood protocol. Each dictionary has the keys:
		- 'reference' : str
			The reference population ('targets' or 'effectors').
		- 'neighbor' : str
			The neighbor population.
		- 'type' : str
			The type of neighborhood ('circle' or 'contact').
		- 'distance' : float
			The distance parameter for the neighborhood.
		- 'description' : str
			The original neighborhood string.

	Notes
	-----
	- The function checks for the existence of pickle files containing target and effector trajectory data.
	- If the files exist, it loads the data and extracts columns that start with 'neighborhood'.
	- The neighborhood settings are extracted using the `extract_neighborhood_settings` function.
	- The function assumes the presence of subdirectories 'output/tables' under the provided `pos`.

	Examples
	--------
	>>> protocols = extract_neighborhoods_from_pickles('/path/to/data')
	>>> for protocol in protocols:
	>>>     print(protocol)
	{'reference': 'targets', 'neighbor': 'targets', 'type': 'contact', 'distance': 5.0, 'description': 'neighborhood_self_contact_5_px'}
	"""

	tab_tc = pos + os.sep.join(['output', 'tables', 'trajectories_targets.pkl'])
	if os.path.exists(tab_tc):
		df_targets = np.load(tab_tc, allow_pickle=True)
	else:
		df_targets = None
	if os.path.exists(tab_tc.replace('targets','effectors')):
		df_effectors = np.load(tab_tc.replace('targets','effectors'), allow_pickle=True)
	else:
		df_effectors = None

	neighborhood_protocols=[]

	if df_targets is not None:
		for column in list(df_targets.columns):
			if column.startswith('neighborhood'):
				neigh_protocol = extract_neighborhood_settings(column, population='targets')
				neighborhood_protocols.append(neigh_protocol)

	if df_effectors is not None:
		for column in list(df_effectors.columns):
			if column.startswith('neighborhood'):
				neigh_protocol = extract_neighborhood_settings(column, population='effectors')
				neighborhood_protocols.append(neigh_protocol)

	return neighborhood_protocols

def extract_neighborhood_settings(neigh_string, population='targets'):
	
	"""
	Extract neighborhood settings from a given string.

	Parameters
	----------
	neigh_string : str
		The string describing the neighborhood settings. Must start with 'neighborhood'.
	population : str, optional
		The population type ('targets' by default). Can be either 'targets' or 'effectors'.

	Returns
	-------
	dict
		A dictionary containing the neighborhood protocol with keys:
		- 'reference' : str
			The reference population.
		- 'neighbor' : str
			The neighbor population.
		- 'type' : str
			The type of neighborhood ('circle' or 'contact').
		- 'distance' : float
			The distance parameter for the neighborhood.
		- 'description' : str
			The original neighborhood string.

	Raises
	------
	AssertionError
		If the `neigh_string` does not start with 'neighborhood'.

	Notes
	-----
	- The function determines the neighbor population based on the given population.
	- The neighborhood type and distance are extracted from the `neigh_string`.
	- The description field in the returned dictionary contains the original neighborhood string.

	Examples
	--------
	>>> extract_neighborhood_settings('neighborhood_self_contact_5_px', 'targets')
	{'reference': 'targets', 'neighbor': 'targets', 'type': 'contact', 'distance': 5.0, 'description': 'neighborhood_self_contact_5_px'}
	"""

	assert neigh_string.startswith('neighborhood')
	if population=='targets':
		neighbor_population = 'effectors'
	elif population=='effectors':
		neighbor_population = 'targets'

	if 'self' in neigh_string:
		
		if 'circle' in neigh_string:
			
			distance = float(neigh_string.split('circle_')[1].replace('_px',''))
			neigh_protocol = {'reference': population,'neighbor': population,'type':'circle','distance':distance,'description': neigh_string}
		elif 'contact' in neigh_string:
			distance=float(neigh_string.split('contact_')[1].replace('_px',''))
			neigh_protocol = {'reference': population,'neighbor': population,'type':'contact','distance':distance,'description': neigh_string}
	else:

		if 'circle' in neigh_string:

			distance=float(neigh_string.split('circle_')[1].replace('_px',''))
			neigh_protocol = {'reference': population,'neighbor': neighbor_population,'type':'circle','distance':distance,'description': neigh_string}
		elif 'contact' in neigh_string:
			
			distance=float(neigh_string.split('contact_')[1].replace('_px',''))
			neigh_protocol = {'reference': population,'neighbor': neighbor_population,'type':'contact','distance':distance,'description': neigh_string}

	return neigh_protocol






from PyQt5.QtWidgets import QWidget, QGridLayout, QVBoxLayout, QLabel, QLineEdit, QHBoxLayout, QRadioButton, QFileDialog, QPushButton, QMessageBox
from PyQt5.QtCore import Qt, QSize
from celldetective.gui.gui_utils import center_window
from celldetective.gui.layouts import ChannelNormGenerator
from celldetective.gui import ThresholdConfigWizard
from PyQt5.QtGui import QDoubleValidator
from superqt.fonticon import icon
from fonticon_mdi6 import MDI6
import numpy as np
from glob import glob
import os
import json
import shutil
from celldetective.gui import Styles
import gc
from cellpose.models import CellposeModel


class SegmentationModelLoader(QWidget, Styles):
	
	"""
	Upload a segmentation model or define a Threshold pipeline.
	"""

	def __init__(self, parent_window):
		
		super().__init__()
		self.parent_window = parent_window
		self.mode = self.parent_window.mode
		if self.mode=="targets":
			self.target_folder = "segmentation_targets"
		elif self.mode=="effectors":
			self.target_folder = "segmentation_effectors"
		self.setWindowTitle('Upload model')
		self.generate_content()
		self.setWindowIcon(self.celldetective_icon)
		center_window(self)
		self.setAttribute(Qt.WA_DeleteOnClose)

	def generate_content(self):

		self.layout = QGridLayout(self)
		self.layout.addWidget(QLabel('Select:'), 0, 0, 1, 1)
		self.layout.setSpacing(5)

		option_layout = QHBoxLayout()
		option_layout.setContentsMargins(10,10,10,10)

		# Create radio buttons
		self.stardist_button = QRadioButton('StarDist')
		#self.stardist_button.setIcon(QIcon(abs_path+f"/icons/star.png"))

		self.cellpose_button = QRadioButton('Cellpose')
		#self.cellpose_button.setIcon(QIcon(abs_path+f"/icons/cellpose.png"))

		self.threshold_button = QRadioButton('Threshold')

		option_layout.addWidget(self.threshold_button)
		option_layout.addWidget(self.stardist_button)
		option_layout.addWidget(self.cellpose_button)

		self.layout.addLayout(option_layout, 1,0,1,2, alignment=Qt.AlignCenter)
		self.generate_base_block()
		self.layout.addLayout(self.base_block, 2,0,1,2)
		# self.combos = self.channel_layout.channel_cbs
		# self.cb_labels = self.channel_layout.channel_labels

		self.generate_cellpose_options()
		self.layout.addLayout(self.cellpose_block, 3,0,1,2)
		self.generate_threshold_options()

		# Create file dialog
		self.file_dialog = QFileDialog()

		# Create button to open file dialog
		self.open_dialog_button = QPushButton('Choose File')
		self.open_dialog_button.clicked.connect(self.showDialog)
		self.file_label = QLabel('No file chosen', self)
		self.layout.addWidget(self.open_dialog_button, 9, 0, 1, 1)
		self.layout.addWidget(self.file_label, 9, 1, 1, 1)

		self.upload_button = QPushButton("Upload")
		self.upload_button.clicked.connect(self.upload_model)
		self.upload_button.setIcon(icon(MDI6.upload,color="white"))
		self.upload_button.setIconSize(QSize(25, 25))
		self.upload_button.setStyleSheet(self.button_style_sheet)
		self.upload_button.setEnabled(False)
		self.layout.addWidget(self.upload_button, 10, 0, 1, 1)
		
		self.base_block_options = [self.calibration_label, self.spatial_calib_le,*self.channel_layout.channel_cbs,*self.channel_layout.channel_labels,*self.channel_layout.normalization_mode_btns, *self.channel_layout.normalization_clip_btns, *self.channel_layout.normalization_min_value_lbl,
								   *self.channel_layout.normalization_min_value_le, *self.channel_layout.normalization_max_value_lbl,*self.channel_layout.normalization_max_value_le,
								   self.channel_layout.add_col_btn]

		self.stardist_button.toggled.connect(self.show_seg_options)
		self.cellpose_button.toggled.connect(self.show_seg_options)
		self.threshold_button.toggled.connect(self.show_seg_options)

		for cb in self.channel_layout.channel_cbs:
			cb.activated.connect(self.unlock_upload)

		self.setLayout(self.layout)

		self.threshold_button.setChecked(True)
		self.show()

	def unlock_upload(self):
		if self.stardist_button.isChecked():
			if np.any([c.currentText()!='--' for c in self.channel_layout.channel_cbs]):
				self.upload_button.setEnabled(True)
			else:
				self.upload_button.setEnabled(False)
		elif self.cellpose_button.isChecked():
			if np.any([c.currentText()!='--' for c in self.channel_layout.channel_cbs]):
				self.upload_button.setEnabled(True)
			else:
				self.upload_button.setEnabled(False)

	def generate_base_block(self):

		"""
		Create widgets common to StarDist and Cellpose.
		"""
		
		self.base_block = QVBoxLayout()
		
		pixel_calib_layout = QHBoxLayout()
		self.calibration_label = QLabel("input spatial\ncalibration: ")
		self.spatial_calib_le = QLineEdit("0,1")
		self.qdv = QDoubleValidator(0.0, np.amax([self.parent_window.parent_window.shape_x, self.parent_window.parent_window.shape_y]), 8, notation=QDoubleValidator.StandardNotation)
		self.spatial_calib_le.setValidator(self.qdv)
		pixel_calib_layout.addWidget(self.calibration_label, 30)
		pixel_calib_layout.addWidget(self.spatial_calib_le, 70)
		self.base_block.addLayout(pixel_calib_layout)

		self.channel_layout = ChannelNormGenerator(self, mode='channels',init_n_channels=2)
		self.channel_layout.setContentsMargins(0,0,0,0)
		self.base_block.addLayout(self.channel_layout)


	def generate_cellpose_options(self):

		"""
		Create Cellpose specific fields to use the model properly when calling it from the app.
		"""

		self.cellpose_block = QVBoxLayout()
		self.cp_diameter_label = QLabel('cell\ndiameter:   ')
		self.cp_diameter_le = QLineEdit("30,0")
		self.cp_diameter_le.setValidator(self.qdv)
		diam_hbox = QHBoxLayout()
		diam_hbox.addWidget(self.cp_diameter_label, 30)
		diam_hbox.addWidget(self.cp_diameter_le, 70)
		self.cellpose_block.addLayout(diam_hbox)

		qdv_prob = QDoubleValidator(-6, 6, 8, notation=QDoubleValidator.StandardNotation)
		self.cp_cellprob_label = QLabel('cellprob\nthreshold:   ')
		self.cp_cellprob_le = QLineEdit('0,0')
		self.cp_cellprob_le.setValidator(qdv_prob)
		cellprob_hbox = QHBoxLayout()
		cellprob_hbox.addWidget(self.cp_cellprob_label, 30)
		cellprob_hbox.addWidget(self.cp_cellprob_le, 70)
		self.cellpose_block.addLayout(cellprob_hbox)

		self.cp_flow_label = QLabel('flow threshold:   ')
		self.cp_flow_le = QLineEdit('0,4')
		self.cp_flow_le.setValidator(qdv_prob)
		flow_hbox = QHBoxLayout()
		flow_hbox.addWidget(self.cp_flow_label, 30)
		flow_hbox.addWidget(self.cp_flow_le, 70)
		self.cellpose_block.addLayout(flow_hbox)

		self.cellpose_options = [self.cp_diameter_label,self.cp_diameter_le,self.cp_cellprob_label,
			self.cp_cellprob_le, self.cp_flow_label, self.cp_flow_le]
		for c in self.cellpose_options:
			c.hide()

	def generate_threshold_options(self):

		"""
		Show the threshold config pipeline button.
		"""

		self.threshold_config_button = QPushButton("Threshold Config Wizard")
		self.threshold_config_button.setIcon(icon(MDI6.auto_fix,color="#1565c0"))
		self.threshold_config_button.setIconSize(QSize(20, 20))
		self.threshold_config_button.setVisible(False)
		self.threshold_config_button.setStyleSheet(self.button_style_sheet_2)
		self.threshold_config_button.clicked.connect(self.open_threshold_config_wizard)
		self.layout.addWidget(self.threshold_config_button,3,0,1,2)
		self.threshold_config_button.hide()

	def showDialog(self):

		"""
		Import the model in the proper folder.
		"""

		# Set the file dialog according to the option chosen
		if self.stardist_button.isChecked():
			self.seg_mode = "stardist"
			self.file_dialog.setFileMode(QFileDialog.Directory)
		elif self.cellpose_button.isChecked():
			self.seg_mode = "cellpose"
			self.file_dialog.setFileMode(QFileDialog.ExistingFile)
		elif self.threshold_button.isChecked():
			self.seg_mode = "threshold"
			self.file_dialog.setFileMode(QFileDialog.ExistingFile)

		# If accepted check validity of data
		if self.file_dialog.exec_() == QFileDialog.Accepted:
			self.filename = self.file_dialog.selectedFiles()[0]
			if self.seg_mode=="stardist":
				subfiles = glob(self.filename+"/*")
				subfiles = [s.replace('\\','/') for s in subfiles]
				if self.filename+"/thresholds.json" in subfiles:
					self.file_label.setText(self.filename.split("/")[-1])
					self.modelname = self.filename.split("/")[-1]
					self.destination = os.path.split(os.path.dirname(os.path.realpath(__file__)))[0]+f"/models/{self.target_folder}/"+self.modelname
					self.folder_dest = self.destination
				else:
					msgBox = QMessageBox()
					msgBox.setIcon(QMessageBox.Warning)
					msgBox.setText("StarDist model not recognized... Please ensure that it contains a thresholds.json file or that it is a valid StarDist model...")
					msgBox.setWindowTitle("Warning")
					msgBox.setStandardButtons(QMessageBox.Ok)
					returnValue = msgBox.exec()
					if returnValue == QMessageBox.Ok:
						return None					
			
			if self.seg_mode=="cellpose":
				self.file_label.setText(self.filename.split("/")[-1])
				self.modelname = self.filename.split("/")[-1]
				print(f"Transferring Cellpose model {self.filename}...")
				self.folder_dest = os.path.split(os.path.dirname(os.path.realpath(__file__)))[0]+f"/models/{self.target_folder}/"+self.modelname
				self.destination = self.folder_dest+f"/{self.modelname}"

			if self.seg_mode=="threshold":
				self.file_label.setText(self.filename.split("/")[-1])

	def show_seg_options(self):

		"""
		Show the relevant widgets, mask the others.
		"""
		self.base_block_options = [self.calibration_label, self.spatial_calib_le,*self.channel_layout.channel_cbs,*self.channel_layout.channel_labels,*self.channel_layout.normalization_mode_btns, *self.channel_layout.normalization_clip_btns, *self.channel_layout.normalization_min_value_lbl,
								   *self.channel_layout.normalization_min_value_le, *self.channel_layout.normalization_max_value_lbl,*self.channel_layout.normalization_max_value_le,
								   self.channel_layout.add_col_btn]

		if self.cellpose_button.isChecked():
			self.spatial_calib_le.setToolTip('Cellpose rescales the images such that the cells are 30.0 pixels. You can compute the scale from the training data as\n(pixel calibration [µm] in training images)*(cell diameter [px] in training images)/(30 [px]).\nIf you pass images with a different calibration to the model, they will be rescaled automatically.\nThe rescaling is ignored if you pass a diameter different from 30 px below.')
			self.channel_layout.channel_labels[0].setText('cyto: ')
			self.channel_layout.channel_labels[1].setText('nuclei: ')
			for c in [self.threshold_config_button]:
				c.hide()
			for c in self.cellpose_options+self.base_block_options:
				c.show()
			self.unlock_upload()
		elif self.stardist_button.isChecked():
			self.spatial_calib_le.setToolTip('')
			self.channel_layout.channel_labels[0].setText('channel 1: ')
			self.channel_layout.channel_labels[1].setText('channel 2: ')
			for c in self.base_block_options:
				c.show()
			for c in self.cellpose_options+[self.threshold_config_button]:
				c.hide()
			self.unlock_upload()
		else:
			for c in self.base_block_options + self.cellpose_options:
				c.hide()
			self.threshold_config_button.show()
			self.upload_button.setEnabled(True)

		self.resize(self.sizeHint());
		#self.adjustSize()

	def upload_model(self):

		"""
		Upload the model.
		"""
		
		channels = []
		for i in range(len(self.channel_layout.channel_cbs)):
			channels.append(self.channel_layout.channel_cbs[i].currentText())

		if self.file_label.text()=='No file chosen':
			msgBox = QMessageBox()
			msgBox.setIcon(QMessageBox.Warning)
			msgBox.setText("Please select a model first.")
			msgBox.setWindowTitle("Warning")
			msgBox.setStandardButtons(QMessageBox.Ok)
			returnValue = msgBox.exec()
			if returnValue == QMessageBox.Ok:
				return None

		if not self.threshold_button.isChecked():
			
			if self.stardist_button.isChecked():
				
				try:
					shutil.copytree(self.filename, self.destination)
				except FileExistsError:
					msgBox = QMessageBox()
					msgBox.setIcon(QMessageBox.Warning)
					msgBox.setText("A model with the same name already exists in the models folder. Please rename it.")
					msgBox.setWindowTitle("Warning")
					msgBox.setStandardButtons(QMessageBox.Ok)
					returnValue = msgBox.exec()
					if returnValue == QMessageBox.Ok:
						return None

			elif self.cellpose_button.isChecked():

				try:
					if not os.path.exists(self.folder_dest):
						os.mkdir(self.folder_dest)
					shutil.copy(self.filename, self.destination)
				except FileExistsError:
					msgBox = QMessageBox()
					msgBox.setIcon(QMessageBox.Warning)
					msgBox.setText("A model with the same name already exists in the models folder. Please rename it.")
					msgBox.setWindowTitle("Warning")
					msgBox.setStandardButtons(QMessageBox.Ok)
					returnValue = msgBox.exec()
					if returnValue == QMessageBox.Ok:
						return None
				
				try:
					model = CellposeModel(pretrained_model=self.destination, model_type=None, nchan=len(channels))
					self.scale_model = model.diam_mean
					print(f'{self.scale_model=}')
					del model
					gc.collect()
				except Exception as e:
					print(e)
					msgBox = QMessageBox()
					msgBox.setIcon(QMessageBox.Critical)
					msgBox.setText(f"Cellpose model could not be loaded...")
					msgBox.setWindowTitle("Error")
					msgBox.setStandardButtons(QMessageBox.Ok)
					returnValue = msgBox.exec()
					if returnValue == QMessageBox.Ok:
						return None					

			self.generate_input_config()
			self.parent_window.init_seg_model_list()
			self.close()
		else:
			if self.mode=="targets":	
				self.parent_window.threshold_config_targets = self.filename
				self.parent_window.seg_model_list.setCurrentText('Threshold')
				print('Path to the traditional segmentation pipeline successfully set in celldetective...')
				self.close()
			elif self.mode=="effectors":
				self.parent_window.threshold_config_effectors = self.filename
				self.parent_window.seg_model_list.setCurrentText('Threshold')
				print('Path to the traditional segmentation pipeline successfully set in celldetective...')
				self.close()	

	def generate_input_config(self):

		"""
		Check the ticked options and input parameters to create
		a configuration to use the uploaded model properly.
		"""
		
		if self.threshold_button.isChecked():
			model_type = "threshold"
			return None

		channels = []
		for i in range(len(self.channel_layout.channel_cbs)):
			channels.append(self.channel_layout.channel_cbs[i].currentText())

		slots_to_keep = np.where(np.array(channels)!='--')[0]
		while '--' in channels:
			channels.remove('--')

		norm_values = np.array([[float(a.replace(',','.')),float(b.replace(',','.'))] for a,b in zip([l.text() for l in self.channel_layout.normalization_min_value_le],
											[l.text() for l in self.channel_layout.normalization_max_value_le])])
		norm_values = norm_values[slots_to_keep]
		norm_values = [list(v) for v in norm_values]

		clip_values = np.array(self.channel_layout.clip_option)
		clip_values = list(clip_values[slots_to_keep])
		clip_values = [bool(c) for c in clip_values]

		normalization_mode = np.array(self.channel_layout.normalization_mode)
		normalization_mode = list(normalization_mode[slots_to_keep])
		normalization_mode = [bool(m) for m in normalization_mode]

		dico = {}

		# Check model option
		if self.stardist_button.isChecked():
			
			# Get channels
			# channels = []
			# for c in self.combos:
			# 	if c.currentText()!="--":
			# 		channels.append(c.currentText())
			model_type = "stardist"
			spatial_calib = float(self.spatial_calib_le.text().replace(',','.'))
			#normalize = self.normalize_checkbox.isChecked()
			dico.update({"channels": channels,
						 #"normalize": normalize,
						 "spatial_calibration": spatial_calib,
						 'normalization_percentile': normalization_mode,
						 'normalization_clip': clip_values,
						 'normalization_values': norm_values,
						 })

		elif self.cellpose_button.isChecked():
			
			# Get channels (cyto and nucleus)
			# channels = []
			# for c in self.combos[:2]:
			# 	if c.currentText()!="--":
			# 		channels.append(c.currentText())

			model_type = "cellpose"
			#cellpose requires at least two channels
			if len(channels)==1:
				channels += ['None']
				normalization_mode += [True]
				clip_values += [True]
				norm_values += [[1,99]]

			diameter = float(self.cp_diameter_le.text().replace(',','.'))
			cellprob_threshold = float(self.cp_cellprob_le.text().replace(',','.'))
			flow_threshold = float(self.cp_flow_le.text().replace(',','.'))
			#normalize = self.normalize_checkbox.isChecked()
			spatial_calib = float(self.spatial_calib_le.text().replace(',','.'))
			# assume 30 px diameter in cellpose model
			spatial_calib = spatial_calib * diameter / self.scale_model
			diameter = 30.0

			dico.update({"channels": channels,
						 "diameter": diameter,
						 "cellprob_threshold": cellprob_threshold,
						 "flow_threshold": flow_threshold,
						 #"normalize": normalize,
						 "spatial_calibration": spatial_calib,
						 'normalization_percentile': normalization_mode,
						 'normalization_clip': clip_values,
						 'normalization_values': norm_values,
						 })


		dico.update({"model_type": model_type})
		print(f"{dico=}")
		json_object = json.dumps(dico, indent=4)

		# Writing to sample.json
		# if os.path.exists(self.folder_dest):

		# 	msgBox = QMessageBox()
		# 	msgBox.setIcon(QMessageBox.Warning)
		# 	msgBox.setText("A model with the same name already exists. Do you want to replace it?")
		# 	msgBox.setWindowTitle("Confirm")
		# 	msgBox.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
		# 	returnValue = msgBox.exec()
		# 	if returnValue == QMessageBox.No:
		# 		return None
		# 	elif returnValue == QMessageBox.Yes:
		# 		shutil.rmtree(self.folder_dest)
		#os.mkdir(self.folder_dest)

		print("Configuration successfully written in ",self.folder_dest+os.sep+"config_input.json")
		with open(self.folder_dest+os.sep+"config_input.json", "w") as outfile:
			outfile.write(json_object)

	def open_threshold_config_wizard(self):

		self.parent_window.parent_window.locate_image()
		if self.parent_window.parent_window.current_stack is None:
			return None
		else:
			self.ThreshWizard = ThresholdConfigWizard(self)
			self.ThreshWizard.show()



"""
Copyright © 2022 Laboratoire Adhesion et Inflammation, Authored by Remy Torro.
"""

import argparse
import datetime
import os
from art import tprint
from celldetective.signals import analyze_signals
import pandas as pd

tprint("Signals")

parser = argparse.ArgumentParser(description="Classify and regress the signals based on the provided model.",
								formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument('-p',"--position", required=True, help="Path to the position")
parser.add_argument('-m',"--model", required=True, help="Path to the model")
parser.add_argument("--mode", default="target", choices=["target","effector","targets","effectors"],help="Cell population of interest")
parser.add_argument("--use_gpu", default="True", choices=["True","False"],help="use GPU")

args = parser.parse_args()
process_arguments = vars(args)
pos = str(process_arguments['position'])
model = str(process_arguments['model'])
mode = str(process_arguments['mode'])
use_gpu = process_arguments['use_gpu']
if use_gpu=='True' or use_gpu=='true' or use_gpu=='1':
	use_gpu = True
else:
	use_gpu = False

column_labels = {'track': "TRACK_ID", 'time': 'FRAME', 'x': 'POSITION_X', 'y': 'POSITION_Y'}

if mode.lower()=="target" or mode.lower()=="targets":
	table_name = "trajectories_targets.csv"

elif mode.lower()=="effector" or mode.lower()=="effectors":
	table_name = "trajectories_effectors.csv"

# Load trajectories, add centroid if not in trajectory
trajectories = pos+os.sep.join(['output','tables', table_name])
if os.path.exists(trajectories):
	trajectories = pd.read_csv(trajectories)
else:
	print('The trajectories table could not be found. Abort.')
	os.abort()

log=f'segmentation model: {model} \n'

with open(pos+f'log_{mode}.json', 'a') as f:
	f.write(f'{datetime.datetime.now()} SIGNAL ANALYSIS \n')
	f.write(log)

trajectories = analyze_signals(trajectories.copy(), model, interpolate_na=True, selected_signals=None, column_labels = column_labels, plot_outcome=True,output_dir=pos+'output/')
trajectories = trajectories.sort_values(by=[column_labels['track'], column_labels['time']])
trajectories.to_csv(pos+os.sep.join(['output','tables', table_name]), index=False)




from PyQt5.QtWidgets import QRadioButton, QButtonGroup, QMainWindow, QTableView, QAction, QMenu,QFileDialog, QLineEdit, QHBoxLayout, QWidget, QPushButton, QVBoxLayout, QComboBox, QLabel, QCheckBox, QMessageBox
from PyQt5.QtCore import Qt, QAbstractTableModel
import pandas as pd
import matplotlib.pyplot as plt
plt.rcParams['svg.fonttype'] = 'none'
from celldetective.gui.gui_utils import FigureCanvas, center_window
from celldetective.utils import differentiate_per_track, collapse_trajectories_by_status
import numpy as np
import seaborn as sns
import matplotlib.cm as mcm
import os
from celldetective.gui import Styles
from superqt import QColormapComboBox, QLabeledSlider, QSearchableComboBox
from superqt.fonticon import icon
from fonticon_mdi6 import MDI6
from math import floor

from matplotlib import colormaps

class PandasModel(QAbstractTableModel):

	"""
	from https://stackoverflow.com/questions/31475965/fastest-way-to-populate-qtableview-from-pandas-data-frame
	"""

	def __init__(self, data):
		QAbstractTableModel.__init__(self)
		self._data = data

	def rowCount(self, parent=None):
		return self._data.shape[0]

	def columnCount(self, parent=None):
		return self._data.shape[1]

	def data(self, index, role=Qt.DisplayRole):
		if index.isValid():
			if role == Qt.DisplayRole:
				return str(self._data.iloc[index.row(), index.column()])
		return None

	def headerData(self, col, orientation, role):
		if orientation == Qt.Horizontal and role == Qt.DisplayRole:
			return self._data.columns[col]
		return None


class QueryWidget(QWidget):

	def __init__(self, parent_window):

		super().__init__()
		self.parent_window = parent_window
		self.setWindowTitle("Filter table")
		# Create the QComboBox and add some items
		center_window(self)

		
		layout = QHBoxLayout(self)
		layout.setContentsMargins(30,30,30,30)
		self.query_le = QLineEdit()
		layout.addWidget(self.query_le, 70)

		self.submit_btn = QPushButton('submit')
		self.submit_btn.clicked.connect(self.filter_table)
		layout.addWidget(self.submit_btn, 30)
		self.setAttribute(Qt.WA_DeleteOnClose)

	def filter_table(self):
		try:
			query_text = self.query_le.text() #.replace('class', '`class`')
			tab = self.parent_window.data.query(query_text)
			self.subtable = TableUI(tab, query_text, plot_mode="static", population=self.parent_window.population)
			self.subtable.show()
			self.close()
		except Exception as e:
			print(e)
			return None


class MergeOneHotWidget(QWidget, Styles):

	def __init__(self, parent_window, selected_columns=None):

		super().__init__()
		self.parent_window = parent_window
		self.selected_columns = selected_columns

		self.setWindowTitle("Merge one-hot encoded columns...")
		# Create the QComboBox and add some items
		center_window(self)
		
		self.layout = QVBoxLayout(self)
		self.layout.setContentsMargins(30,30,30,30)

		if self.selected_columns is not None:
			n_cols = len(self.selected_columns)
		else:
			n_cols = 2

		name_hbox = QHBoxLayout()
		name_hbox.addWidget(QLabel('New categorical column: '), 33)
		self.new_col_le = QLineEdit()
		self.new_col_le.setText('categorical_')
		self.new_col_le.textChanged.connect(self.allow_merge)
		name_hbox.addWidget(self.new_col_le, 66)
		self.layout.addLayout(name_hbox)


		self.layout.addWidget(QLabel('Source columns: '))

		self.cbs = [QSearchableComboBox() for i in range(n_cols)]
		self.cbs_layout = QVBoxLayout()

		for i in range(n_cols):
			lay = QHBoxLayout()
			lay.addWidget(QLabel(f'column {i}: '), 33)
			self.cbs[i].addItems(['--']+list(self.parent_window.data.columns))
			if self.selected_columns is not None:
				self.cbs[i].setCurrentText(self.selected_columns[i])
			lay.addWidget(self.cbs[i], 66)
			self.cbs_layout.addLayout(lay)

		self.layout.addLayout(self.cbs_layout)

		hbox = QHBoxLayout()
		self.add_col_btn = QPushButton('Add column')
		self.add_col_btn.clicked.connect(self.add_col)
		self.add_col_btn.setStyleSheet(self.button_add)
		self.add_col_btn.setIcon(icon(MDI6.plus,color="black"))

		hbox.addWidget(QLabel(''), 50)
		hbox.addWidget(self.add_col_btn, 50, alignment=Qt.AlignRight)
		self.layout.addLayout(hbox)

		self.submit_btn = QPushButton('Merge')
		self.submit_btn.setStyleSheet(self.button_style_sheet)
		self.submit_btn.clicked.connect(self.merge_cols)
		self.layout.addWidget(self.submit_btn, 30)

		self.setAttribute(Qt.WA_DeleteOnClose)

	def add_col(self):
		self.cbs.append(QSearchableComboBox())
		self.cbs[-1].addItems(['--']+list(self.parent_window.data.columns))
		lay = QHBoxLayout()
		lay.addWidget(QLabel(f'column {len(self.cbs)-1}: '), 33)
		lay.addWidget(self.cbs[-1], 66)
		self.cbs_layout.addLayout(lay)	

	def merge_cols(self):
		
		self.parent_window.data[self.new_col_le.text()] = self.parent_window.data.loc[:,list(self.selected_columns)].idxmax(axis=1)
		self.parent_window.model = PandasModel(self.parent_window.data)
		self.parent_window.table_view.setModel(self.parent_window.model)
		self.close()

	def allow_merge(self):

		if self.new_col_le.text()=='':
			self.submit_btn.setEnabled(False)
		else:
			self.submit_btn.setEnabled(True)


class DifferentiateColWidget(QWidget, Styles):

	def __init__(self, parent_window, column=None):

		super().__init__()
		self.parent_window = parent_window
		self.column = column

		self.setWindowTitle("d/dt")
		# Create the QComboBox and add some items
		center_window(self)
		
		layout = QVBoxLayout(self)
		layout.setContentsMargins(30,30,30,30)

		self.measurements_cb = QComboBox()
		self.measurements_cb.addItems(list(self.parent_window.data.columns))
		if self.column is not None:
			idx = self.measurements_cb.findText(self.column)
			self.measurements_cb.setCurrentIndex(idx)

		measurement_layout = QHBoxLayout()
		measurement_layout.addWidget(QLabel('measurements: '), 25)
		measurement_layout.addWidget(self.measurements_cb, 75)
		layout.addLayout(measurement_layout)

		self.window_size_slider = QLabeledSlider()
		self.window_size_slider.setRange(1,np.nanmax(self.parent_window.data.FRAME.to_numpy()))
		self.window_size_slider.setValue(3)
		window_layout = QHBoxLayout()
		window_layout.addWidget(QLabel('window size: '), 25)
		window_layout.addWidget(self.window_size_slider, 75)
		layout.addLayout(window_layout)

		self.backward_btn = QRadioButton('backward')
		self.bi_btn = QRadioButton('bi')
		self.bi_btn.click()
		self.forward_btn = QRadioButton('forward')
		self.mode_btn_group = QButtonGroup()
		self.mode_btn_group.addButton(self.backward_btn)
		self.mode_btn_group.addButton(self.bi_btn)
		self.mode_btn_group.addButton(self.forward_btn)

		mode_layout = QHBoxLayout()
		mode_layout.addWidget(QLabel('mode: '),25)
		mode_sublayout = QHBoxLayout()
		mode_sublayout.addWidget(self.backward_btn, 33, alignment=Qt.AlignCenter)
		mode_sublayout.addWidget(self.bi_btn, 33, alignment=Qt.AlignCenter)
		mode_sublayout.addWidget(self.forward_btn, 33, alignment=Qt.AlignCenter)
		mode_layout.addLayout(mode_sublayout, 75)
		layout.addLayout(mode_layout)

		self.submit_btn = QPushButton('Compute')
		self.submit_btn.setStyleSheet(self.button_style_sheet)
		self.submit_btn.clicked.connect(self.compute_derivative_and_add_new_column)
		layout.addWidget(self.submit_btn, 30)

		self.setAttribute(Qt.WA_DeleteOnClose)


	def compute_derivative_and_add_new_column(self):
		
		if self.bi_btn.isChecked():
			mode = 'bi'
		elif self.forward_btn.isChecked():
			mode = 'forward'
		elif self.backward_btn.isChecked():
			mode = 'backward'
		self.parent_window.data = differentiate_per_track(self.parent_window.data,
														  self.measurements_cb.currentText(),
														  window_size=self.window_size_slider.value(),
														  mode=mode)
		self.parent_window.model = PandasModel(self.parent_window.data)
		self.parent_window.table_view.setModel(self.parent_window.model)
		self.close()

class AbsColWidget(QWidget, Styles):

	def __init__(self, parent_window, column=None):

		super().__init__()
		self.parent_window = parent_window
		self.column = column

		self.setWindowTitle("abs(.)")
		# Create the QComboBox and add some items
		center_window(self)
		
		layout = QVBoxLayout(self)
		layout.setContentsMargins(30,30,30,30)

		self.measurements_cb = QComboBox()
		self.measurements_cb.addItems(list(self.parent_window.data.columns))
		if self.column is not None:
			idx = self.measurements_cb.findText(self.column)
			self.measurements_cb.setCurrentIndex(idx)

		measurement_layout = QHBoxLayout()
		measurement_layout.addWidget(QLabel('measurements: '), 25)
		measurement_layout.addWidget(self.measurements_cb, 75)
		layout.addLayout(measurement_layout)

		self.submit_btn = QPushButton('Compute')
		self.submit_btn.setStyleSheet(self.button_style_sheet)
		self.submit_btn.clicked.connect(self.compute_abs_and_add_new_column)
		layout.addWidget(self.submit_btn, 30)

		self.setAttribute(Qt.WA_DeleteOnClose)


	def compute_abs_and_add_new_column(self):
		

		self.parent_window.data['|'+self.measurements_cb.currentText()+'|'] = self.parent_window.data[self.measurements_cb.currentText()].abs()
		self.parent_window.model = PandasModel(self.parent_window.data)
		self.parent_window.table_view.setModel(self.parent_window.model)
		self.close()

class RenameColWidget(QWidget):

	def __init__(self, parent_window, column=None):

		super().__init__()
		self.parent_window = parent_window
		self.column = column
		if self.column is None:
			self.column = ''

		self.setWindowTitle("Rename column")
		# Create the QComboBox and add some items
		center_window(self)
		
		layout = QHBoxLayout(self)
		layout.setContentsMargins(30,30,30,30)
		self.new_col_name = QLineEdit()
		self.new_col_name.setText(self.column)
		layout.addWidget(self.new_col_name, 70)

		self.submit_btn = QPushButton('rename')
		self.submit_btn.clicked.connect(self.rename_col)
		layout.addWidget(self.submit_btn, 30)
		self.setAttribute(Qt.WA_DeleteOnClose)

	def rename_col(self):
		
		old_name = self.column
		new_name = self.new_col_name.text()
		self.parent_window.data = self.parent_window.data.rename(columns={old_name: new_name})

		self.parent_window.model = PandasModel(self.parent_window.data)
		self.parent_window.table_view.setModel(self.parent_window.model)
		self.close()


class TableUI(QMainWindow, Styles):

	def __init__(self, data, title, population='targets',plot_mode="plot_track_signals", *args, **kwargs):

		QMainWindow.__init__(self, *args, **kwargs)

		self.setWindowTitle(title)
		self.setGeometry(100,100,1000,400)
		center_window(self)
		self.title = title
		self.plot_mode = plot_mode
		self.population = population
		self.numerics = ['int16', 'int32', 'int64', 'float16', 'float32', 'float64']
		self.groupby_cols = ['position', 'TRACK_ID']
		self.tracks = False

		if self.population=='pairs':
			self.groupby_cols = ['position','reference_population', 'neighbor_population','REFERENCE_ID', 'NEIGHBOR_ID']
			self.tracks = True # for now
		else:
			if 'TRACK_ID' in data.columns:
				if not np.all(data['TRACK_ID'].isnull()):
					self.tracks = True

		self._createMenuBar()
		self._createActions()

		self.table_view = QTableView(self)
		self.setCentralWidget(self.table_view)

		# Set the model for the table view
		self.data = data

		self.model = PandasModel(data)
		self.table_view.setModel(self.model)
		self.table_view.resizeColumnsToContents()
		self.setAttribute(Qt.WA_DeleteOnClose)


	def _createActions(self):

			self.save_as = QAction("&Save as...", self)
			self.save_as.triggered.connect(self.save_as_csv)
			self.save_as.setShortcut("Ctrl+s")
			self.fileMenu.addAction(self.save_as)

			self.save_inplace = QAction("&Save inplace...", self)
			self.save_inplace.triggered.connect(self.save_as_csv_inplace_per_pos)
			#self.save_inplace.setShortcut("Ctrl+s")
			self.fileMenu.addAction(self.save_inplace)

			self.plot_action = QAction("&Plot...", self)
			self.plot_action.triggered.connect(self.plot)
			self.plot_action.setShortcut("Ctrl+p")
			self.fileMenu.addAction(self.plot_action)		

			self.plot_inst_action = QAction("&Plot instantaneous...", self)
			self.plot_inst_action.triggered.connect(self.plot_instantaneous)
			self.plot_inst_action.setShortcut("Ctrl+i")
			self.fileMenu.addAction(self.plot_inst_action)	

			self.groupby_action = QAction("&Group by tracks...", self)
			self.groupby_action.triggered.connect(self.set_projection_mode_tracks)
			self.groupby_action.setShortcut("Ctrl+g")
			self.fileMenu.addAction(self.groupby_action)
			if not self.tracks:
				self.groupby_action.setEnabled(False)

			if self.population=='pairs':
				self.groupby_neigh_action = QAction("&Group by neighbors...", self)
				self.groupby_neigh_action.triggered.connect(self.set_projection_mode_neigh)
				self.fileMenu.addAction(self.groupby_neigh_action)	

				self.groupby_ref_action = QAction("&Group by reference...", self)
				self.groupby_ref_action.triggered.connect(self.set_projection_mode_ref)
				self.fileMenu.addAction(self.groupby_ref_action)	


			self.groupby_time_action = QAction("&Group by frames...", self)
			self.groupby_time_action.triggered.connect(self.groupby_time_table)
			self.groupby_time_action.setShortcut("Ctrl+t")
			self.fileMenu.addAction(self.groupby_time_action)

			self.query_action = QAction('Query...', self)
			self.query_action.triggered.connect(self.perform_query)
			self.fileMenu.addAction(self.query_action)

			self.delete_action = QAction('&Delete...', self)
			self.delete_action.triggered.connect(self.delete_columns)
			self.delete_action.setShortcut(Qt.Key_Delete)
			self.editMenu.addAction(self.delete_action)

			self.rename_col_action = QAction('&Rename...', self)
			self.rename_col_action.triggered.connect(self.rename_column)
			#self.rename_col_action.setShortcut(Qt.Key_Delete)
			self.editMenu.addAction(self.rename_col_action)

			if self.population=='pairs':
				self.merge_action = QAction('&Merge...', self)
				self.merge_action.triggered.connect(self.merge_tables)
				#self.rename_col_action.setShortcut(Qt.Key_Delete)
				self.editMenu.addAction(self.merge_action)

			self.derivative_action = QAction('&Differentiate...', self)
			self.derivative_action.triggered.connect(self.differenciate_selected_feature)
			self.derivative_action.setShortcut("Ctrl+D")
			self.mathMenu.addAction(self.derivative_action)

			self.abs_action = QAction('&Absolute value...', self)
			self.abs_action.triggered.connect(self.take_abs_of_selected_feature)
			#self.derivative_action.setShortcut("Ctrl+D")
			self.mathMenu.addAction(self.abs_action)			

			self.onehot_action = QAction('&One hot to categorical...', self)
			self.onehot_action.triggered.connect(self.transform_one_hot_cols_to_categorical)
			#self.onehot_action.setShortcut("Ctrl+D")
			self.mathMenu.addAction(self.onehot_action)		

	def merge_tables(self):

		expanded_table = []
		
		for neigh, group in self.data.groupby(['reference_population','neighbor_population']):
			print(f'{neigh=}')
			ref_pop = neigh[0]; neigh_pop = neigh[1];
			for pos,pos_group in group.groupby('position'):
				print(f'{pos=}')
				
				ref_tab = os.sep.join([pos,'output','tables',f'trajectories_{ref_pop}.csv'])
				neigh_tab = os.sep.join([pos,'output','tables',f'trajectories_{neigh_pop}.csv'])
				if os.path.exists(ref_tab):
					df_ref = pd.read_csv(ref_tab)
					if 'TRACK_ID' in df_ref.columns:
						if not np.all(df_ref['TRACK_ID'].isnull()):
							ref_merge_cols = ['TRACK_ID','FRAME']
						else:
							ref_merge_cols = ['ID','FRAME']
					else:
						ref_merge_cols = ['ID','FRAME']
				if os.path.exists(neigh_tab):
					df_neigh = pd.read_csv(neigh_tab)
					if 'TRACK_ID' in df_neigh.columns:
						if not np.all(df_neigh['TRACK_ID'].isnull()):
							neigh_merge_cols = ['TRACK_ID','FRAME']
						else:
							neigh_merge_cols = ['ID','FRAME']
					else:
						neigh_merge_cols = ['ID','FRAME']

				df_ref = df_ref.add_prefix('reference_',axis=1)
				df_neigh = df_neigh.add_prefix('neighbor_',axis=1)
				ref_merge_cols = ['reference_'+c for c in ref_merge_cols]
				neigh_merge_cols = ['neighbor_'+c for c in neigh_merge_cols]

				merge_ref = pos_group.merge(df_ref, how='outer', left_on=['REFERENCE_ID','FRAME'], right_on=ref_merge_cols, suffixes=('', '_reference'))
				print(f'{merge_ref.columns=}')
				merge_neigh = merge_ref.merge(df_neigh, how='outer', left_on=['NEIGHBOR_ID','FRAME'], right_on=neigh_merge_cols, suffixes=('_reference', '_neighbor'))
				print(f'{merge_neigh.columns=}')
				expanded_table.append(merge_neigh)

		df_expanded = pd.concat(expanded_table, axis=0, ignore_index = True)
		df_expanded = df_expanded.sort_values(by=['position', 'reference_population','neighbor_population','REFERENCE_ID','NEIGHBOR_ID','FRAME'])
		df_expanded = df_expanded.dropna(axis=0, subset=['REFERENCE_ID','NEIGHBOR_ID','reference_population','neighbor_population'])
		self.subtable = TableUI(df_expanded, 'merge', plot_mode = "static", population='pairs')
		self.subtable.show()	


	def delete_columns(self):

		x = self.table_view.selectedIndexes()
		col_idx = np.unique(np.array([l.column() for l in x]))
		cols = np.array(list(self.data.columns))

		msgBox = QMessageBox()
		msgBox.setIcon(QMessageBox.Question)
		msgBox.setText(f"You are about to delete columns {cols[col_idx]}... Do you want to proceed?")
		msgBox.setWindowTitle("Info")
		msgBox.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
		returnValue = msgBox.exec()
		if returnValue == QMessageBox.No:
			return None

		self.data = self.data.drop(list(cols[col_idx]),axis=1)
		self.model = PandasModel(self.data)
		self.table_view.setModel(self.model)

	def rename_column(self):

		x = self.table_view.selectedIndexes()
		col_idx = np.unique(np.array([l.column() for l in x]))

		if len(col_idx) == 0:
			msgBox = QMessageBox()
			msgBox.setIcon(QMessageBox.Question)
			msgBox.setText(f"Please select a column first.")
			msgBox.setWindowTitle("Warning")
			msgBox.setStandardButtons(QMessageBox.Ok)
			returnValue = msgBox.exec()
			if returnValue == QMessageBox.Ok:
				return None
			else:
				return None
			
		cols = np.array(list(self.data.columns))
		selected_col = str(cols[col_idx][0])

		self.renameWidget = RenameColWidget(self, selected_col)
		self.renameWidget.show()

	def save_as_csv_inplace_per_pos(self):

		print("Saving each table in its respective position folder...")
		for pos,pos_group in self.data.groupby('position'):
			pos_group.to_csv(pos+os.sep.join(['output', 'tables', f'trajectories_{self.population}.csv']), index=False)
		print("Done...")

	def differenciate_selected_feature(self):
		
		# check only one col selected and assert is numerical
		# open widget to select window parameters, directionality
		# create new col
		
		x = self.table_view.selectedIndexes()
		col_idx = np.unique(np.array([l.column() for l in x]))
		if col_idx!=0:
			cols = np.array(list(self.data.columns))
			selected_col = str(cols[col_idx][0])
		else:
			selected_col = None

		self.diffWidget = DifferentiateColWidget(self, selected_col)
		self.diffWidget.show()

	def take_abs_of_selected_feature(self):
		
		# check only one col selected and assert is numerical
		# open widget to select window parameters, directionality
		# create new col
		
		x = self.table_view.selectedIndexes()
		col_idx = np.unique(np.array([l.column() for l in x]))
		if col_idx!=0:
			cols = np.array(list(self.data.columns))
			selected_col = str(cols[col_idx][0])
		else:
			selected_col = None

		self.absWidget = AbsColWidget(self, selected_col)
		self.absWidget.show()


	def transform_one_hot_cols_to_categorical(self):

		x = self.table_view.selectedIndexes()
		col_idx = np.unique(np.array([l.column() for l in x]))
		if list(col_idx):
			cols = np.array(list(self.data.columns))
			selected_cols = cols[col_idx]
		else:
			selected_cols = None

		self.mergewidget = MergeOneHotWidget(self, selected_columns=selected_cols)
		self.mergewidget.show()


	def groupby_time_table(self):

		"""
		
		Perform a time average across each track for all features

		"""

		num_df = self.data.select_dtypes(include=self.numerics)

		timeseries = num_df.groupby("FRAME").sum().copy()
		timeseries["timeline"] = timeseries.index
		self.subtable = TableUI(timeseries,"Group by frames", plot_mode="plot_timeseries")
		self.subtable.show()

	def perform_query(self):

		"""
		
		Perform a time average across each track for all features

		"""
		self.query_widget = QueryWidget(self)
		self.query_widget.show()

		# num_df = self.data.select_dtypes(include=self.numerics)

		# timeseries = num_df.groupby("FRAME").mean().copy()
		# timeseries["timeline"] = timeseries.index
		# self.subtable = TableUI(timeseries,"Group by frames", plot_mode="plot_timeseries")
		# self.subtable.show()

	def set_projection_mode_neigh(self):

		self.groupby_cols = ['position', 'reference_population', 'neighbor_population', 'NEIGHBOR_ID', 'FRAME']
		self.set_projection_mode_tracks()

	def set_projection_mode_ref(self):
		
		self.groupby_cols = ['position', 'reference_population', 'neighbor_population', 'REFERENCE_ID', 'FRAME']
		self.set_projection_mode_tracks()

	def set_projection_mode_tracks(self):

		self.projectionWidget = QWidget()
		self.projectionWidget.setMinimumWidth(500)
		self.projectionWidget.setWindowTitle('Set projection mode')
		
		layout = QVBoxLayout()
		self.projectionWidget.setLayout(layout)

		self.projection_option = QRadioButton('global operation: ')
		self.projection_option.setToolTip('Collapse the cell track measurements with an operation over each track.')
		self.projection_option.setChecked(True)
		self.projection_option.toggled.connect(self.enable_projection_options)
		self.projection_op_cb = QComboBox()
		self.projection_op_cb.addItems(['mean','median','min','max', 'prod', 'sum'])

		projection_layout = QHBoxLayout()
		projection_layout.addWidget(self.projection_option, 33)
		projection_layout.addWidget(self.projection_op_cb, 66)
		layout.addLayout(projection_layout)

		self.event_time_option = QRadioButton('@event time: ')
		self.event_time_option.setToolTip('Pick the measurements at a specific event time.')
		self.event_time_option.toggled.connect(self.enable_projection_options)
		self.event_times_cb = QComboBox()
		cols = np.array(self.data.columns)
		time_cols = np.array([c.startswith('t_') for c in cols])
		time_cols = list(cols[time_cols])
		if 't0' in list(self.data.columns):
			time_cols.append('t0')
		self.event_times_cb.addItems(time_cols)
		self.event_times_cb.setEnabled(False)

		event_time_layout = QHBoxLayout()
		event_time_layout.addWidget(self.event_time_option, 33)
		event_time_layout.addWidget(self.event_times_cb, 66)
		layout.addLayout(event_time_layout)


		self.per_status_option = QRadioButton('per status: ')
		self.per_status_option.setToolTip('Collapse the cell track measurements independently for each of the cell state.')
		self.per_status_option.toggled.connect(self.enable_projection_options)
		self.per_status_cb = QComboBox()
		self.status_operation = QComboBox()
		self.status_operation.setEnabled(False)
		self.status_operation.addItems(['mean','median','min','max', 'prod', 'sum'])

		status_cols = np.array([c.startswith('status_') or c.startswith('group_') for c in cols])
		status_cols = list(cols[status_cols])
		if 'status' in list(self.data.columns):
			status_cols.append('status')
		self.per_status_cb.addItems(status_cols)
		self.per_status_cb.setEnabled(False)

		per_status_layout = QHBoxLayout()
		per_status_layout.addWidget(self.per_status_option, 33)
		per_status_layout.addWidget(self.per_status_cb, 66)
		layout.addLayout(per_status_layout)

		status_operation_layout = QHBoxLayout()
		status_operation_layout.addWidget(QLabel('operation: '), 33, alignment=Qt.AlignRight)
		status_operation_layout.addWidget(self.status_operation, 66)
		layout.addLayout(status_operation_layout)

		self.btn_projection_group = QButtonGroup()
		self.btn_projection_group.addButton(self.projection_option)
		self.btn_projection_group.addButton(self.event_time_option)
		self.btn_projection_group.addButton(self.per_status_option)

		apply_layout = QHBoxLayout()

		self.set_projection_btn = QPushButton('Apply')
		self.set_projection_btn.setStyleSheet(self.button_style_sheet)
		self.set_projection_btn.clicked.connect(self.set_proj_mode)
		apply_layout.addWidget(QLabel(''), 33)
		apply_layout.addWidget(self.set_projection_btn,33)
		apply_layout.addWidget(QLabel(''),33)
		layout.addLayout(apply_layout)

		self.projectionWidget.show()
		center_window(self.projectionWidget)

	def enable_projection_options(self):

		if self.projection_option.isChecked():
			self.projection_op_cb.setEnabled(True)
			self.event_times_cb.setEnabled(False)
			self.per_status_cb.setEnabled(False)
			self.status_operation.setEnabled(False)
		elif self.event_time_option.isChecked():
			self.projection_op_cb.setEnabled(False)
			self.event_times_cb.setEnabled(True)
			self.per_status_cb.setEnabled(False)
			self.status_operation.setEnabled(False)
		elif self.per_status_option.isChecked():
			self.projection_op_cb.setEnabled(False)
			self.event_times_cb.setEnabled(False)
			self.per_status_cb.setEnabled(True)
			self.status_operation.setEnabled(True)

	def set_1D_plot_params(self):

		self.plot1Dparams = QWidget()
		self.plot1Dparams.setWindowTitle('Set 1D plot parameters')
		
		layout = QVBoxLayout()
		self.plot1Dparams.setLayout(layout)

		layout.addWidget(QLabel('Representations: '))
		self.hist_check = QCheckBox('histogram')
		self.kde_check = QCheckBox('KDE plot')
		self.count_check = QCheckBox('countplot')
		self.ecdf_check = QCheckBox('ECDF plot')
		self.scat_check = QCheckBox('scatter plot')
		self.swarm_check = QCheckBox('swarm')
		self.violin_check = QCheckBox('violin')
		self.strip_check = QCheckBox('strip')
		self.box_check = QCheckBox('boxplot')
		self.boxenplot_check = QCheckBox('boxenplot')

		layout.addWidget(self.hist_check)
		layout.addWidget(self.kde_check)
		layout.addWidget(self.count_check)
		layout.addWidget(self.ecdf_check)
		layout.addWidget(self.scat_check)
		layout.addWidget(self.swarm_check)
		layout.addWidget(self.violin_check)
		layout.addWidget(self.strip_check)
		layout.addWidget(self.box_check)
		layout.addWidget(self.boxenplot_check)

		self.x_cb = QSearchableComboBox()
		self.x_cb.addItems(['--']+list(self.data.columns))

		self.y_cb = QSearchableComboBox()
		self.y_cb.addItems(['--']+list(self.data.columns))

		self.hue_cb = QSearchableComboBox()
		self.hue_cb.addItems(['--']+list(self.data.columns))
		idx = self.hue_cb.findText('--')
		self.hue_cb.setCurrentIndex(idx)

		# Set selected column

		try:
			x = self.table_view.selectedIndexes()
			col_idx = np.array([l.column() for l in x])
			row_idx = np.array([l.row() for l in x])
			column_names = self.data.columns
			unique_cols = np.unique(col_idx)[0]
			y = column_names[unique_cols]
			idx = self.y_cb.findText(y)
			self.y_cb.setCurrentIndex(idx)
		except:
			pass

		hbox = QHBoxLayout()
		hbox.addWidget(QLabel('x: '), 33)
		hbox.addWidget(self.x_cb, 66)
		layout.addLayout(hbox)

		hbox = QHBoxLayout()
		hbox.addWidget(QLabel('y: '), 33)
		hbox.addWidget(self.y_cb, 66)
		layout.addLayout(hbox)

		hbox = QHBoxLayout()
		hbox.addWidget(QLabel('hue: '), 33)
		hbox.addWidget(self.hue_cb, 66)
		layout.addLayout(hbox)

		self.cmap_cb = QColormapComboBox()
		for cm in list(colormaps):
			try:
				self.cmap_cb.addColormap(cm)
			except:
				pass
		
		hbox = QHBoxLayout()
		hbox.addWidget(QLabel('colormap: '), 33)
		hbox.addWidget(self.cmap_cb, 66)
		layout.addLayout(hbox)

		self.plot1d_btn = QPushButton('set')
		self.plot1d_btn.setStyleSheet(self.button_style_sheet)
		self.plot1d_btn.clicked.connect(self.plot1d)
		layout.addWidget(self.plot1d_btn)

		self.plot1Dparams.show()
		center_window(self.plot1Dparams)


	def plot1d(self):

		self.x_option = False
		if self.x_cb.currentText()!='--':
			self.x_option = True
			self.x = self.x_cb.currentText()

		self.fig, self.ax = plt.subplots(1,1,figsize=(4,3))
		self.plot1dWindow = FigureCanvas(self.fig, title="scatter")
		self.ax.clear()

		cmap = getattr(mcm, self.cmap_cb.currentText())

		try:
			hue_variable = self.hue_cb.currentText()
			colors = [cmap(i / len(self.data[hue_variable].unique())) for i in range(len(self.data[hue_variable].unique()))]
		except:
			colors = None

		if self.hue_cb.currentText()=='--':
			hue_variable = None

		if self.y_cb.currentText()=='--':
			self.y = None
		else:
			self.y = self.y_cb.currentText()

		if self.x_cb.currentText()=='--':
			self.x = None
		else:
			self.x = self.x_cb.currentText()

		legend=True
		
		if self.hist_check.isChecked():
			if self.x is not None:
				sns.histplot(data=self.data, x=self.x, hue=hue_variable, legend=legend, ax=self.ax, palette=colors, kde=True, common_norm=False, stat='density')
				legend = False
			elif self.x is None and self.y is not None:
				sns.histplot(data=self.data, x=self.y, hue=hue_variable, legend=legend, ax=self.ax, palette=colors, kde=True, common_norm=False, stat='density')
				legend = False
			else:
				pass

		if self.kde_check.isChecked():
			if self.x is not None:
				sns.kdeplot(data=self.data, x=self.x, hue=hue_variable, legend=legend, ax=self.ax, palette=colors, cut=0)
				legend = False
			elif self.x is None and self.y is not None:
				sns.kdeplot(data=self.data, x=self.y, hue=hue_variable, legend=legend, ax=self.ax, palette=colors, cut=0)
				legend = False	
			else:
				pass

		if self.count_check.isChecked():
			sns.countplot(data=self.data, x=self.x, hue=hue_variable, legend=legend, ax=self.ax, palette=colors)
			legend = False


		if self.ecdf_check.isChecked():
			if self.x is not None:
				sns.ecdfplot(data=self.data, x=self.x, hue=hue_variable, legend=legend, ax=self.ax, palette=colors)
				legend = False
			elif self.x is None and self.y is not None:
				sns.ecdfplot(data=self.data, x=self.y, hue=hue_variable, legend=legend, ax=self.ax, palette=colors)
				legend = False
			else:
				pass
						
		if self.scat_check.isChecked():
			if self.x_option:
				sns.scatterplot(data=self.data, x=self.x,y=self.y, hue=hue_variable,legend=legend, ax=self.ax, palette=colors)
				legend = False
			else:
				print('please provide a -x variable...')
				pass

		if self.swarm_check.isChecked():
			if self.x_option:
				sns.swarmplot(data=self.data, x=self.x,y=self.y,dodge=True, hue=hue_variable,legend=legend, ax=self.ax, palette=colors)
				legend = False
			else:
				sns.swarmplot(data=self.data, y=self.y,dodge=True, hue=hue_variable,legend=legend, ax=self.ax, palette=colors)
				legend = False

		if self.violin_check.isChecked():
			if self.x_option:
				sns.violinplot(data=self.data,x=self.x, y=self.y,dodge=True, ax=self.ax, hue=hue_variable, legend=legend, palette=colors)
				legend = False
			else:
				sns.violinplot(data=self.data, y=self.y,dodge=True, hue=hue_variable,legend=legend, ax=self.ax, palette=colors, cut=0)
				legend = False

		if self.box_check.isChecked():
			if self.x_option:
				sns.boxplot(data=self.data, x=self.x, y=self.y,dodge=True, hue=hue_variable,legend=legend, ax=self.ax, fill=False,palette=colors, linewidth=2,)
				legend = False
			else:
				sns.boxplot(data=self.data, y=self.y,dodge=True, hue=hue_variable,legend=legend, ax=self.ax, fill=False,palette=colors, linewidth=2,)
				legend = False

		if self.boxenplot_check.isChecked():
			if self.x_option:
				sns.boxenplot(data=self.data, x=self.x, y=self.y,dodge=True, hue=hue_variable,legend=legend, ax=self.ax, fill=False,palette=colors, linewidth=2,)
				legend = False
			else:
				sns.boxenplot(data=self.data, y=self.y,dodge=True, hue=hue_variable,legend=legend, ax=self.ax, fill=False,palette=colors, linewidth=2,)
				legend = False

		if self.strip_check.isChecked():
			if self.x_option:
				sns.stripplot(data=self.data, x = self.x, y=self.y,dodge=True, ax=self.ax, hue=hue_variable, legend=legend, palette=colors)
				legend = False
			else:
				sns.stripplot(data=self.data, y=self.y,dodge=True, ax=self.ax, hue=hue_variable, legend=legend, palette=colors)
				legend = False

		plt.tight_layout()
		self.fig.set_facecolor('none')  # or 'None'
		self.fig.canvas.setStyleSheet("background-color: transparent;")
		self.plot1dWindow.canvas.draw()
		self.plot1dWindow.show()


	def set_proj_mode(self):
		
		self.static_columns = ['well_index', 'well_name', 'pos_name', 'position', 'well', 'status', 't0', 'class','cell_type','concentration', 'antibody', 'pharmaceutical_agent','TRACK_ID','position', 'neighbor_population', 'reference_population', 'NEIGHBOR_ID', 'REFERENCE_ID', 'FRAME']

		if self.projection_option.isChecked():

			self.projection_mode = self.projection_op_cb.currentText()
			op = getattr(self.data.groupby(self.groupby_cols), self.projection_mode)
			group_table = op(self.data.groupby(self.groupby_cols))

			for c in self.static_columns:
				try:
					group_table[c] = self.data.groupby(self.groupby_cols)[c].apply(lambda x: x.unique()[0])
				except Exception as e:
					print(e)
					pass
			
			if self.population=='pairs':
				for col in reversed(self.groupby_cols): #['neighbor_population', 'reference_population', 'NEIGHBOR_ID', 'REFERENCE_ID']
					if col in group_table:
						first_column = group_table.pop(col)
						group_table.insert(0, col, first_column)				
			else:
				for col in ['TRACK_ID']:
					first_column = group_table.pop(col) 
					group_table.insert(0, col, first_column)
				group_table.pop('FRAME')


		elif self.event_time_option.isChecked():

			time_of_interest = self.event_times_cb.currentText()
			self.projection_mode = f"measurements at {time_of_interest}"
			new_table = []
			for tid,group in self.data.groupby(self.groupby_cols):
				time = group[time_of_interest].values[0]
				if time==time:
					time = floor(time) # floor for onset
				else:
					continue
				frames = group['FRAME'].values
				values = group.loc[group['FRAME']==time,:].to_numpy()
				if len(values)>0:
					values = dict(zip(list(self.data.columns), values[0]))
					for k,c in enumerate(self.groupby_cols):
						values.update({c: tid[k]})
					new_table.append(values)
			
			group_table = pd.DataFrame(new_table)
			if self.population=='pairs':
				for col in self.groupby_cols[1:]:
					first_column = group_table.pop(col) 
					group_table.insert(0, col, first_column)				
			else:
				for col in ['TRACK_ID']:
					first_column = group_table.pop(col) 
					group_table.insert(0, col, first_column)
				
			group_table = group_table.sort_values(by=self.groupby_cols+['FRAME'],ignore_index=True)
			group_table = group_table.reset_index(drop=True)


		elif self.per_status_option.isChecked():

			group_table = collapse_trajectories_by_status(self.data, status=self.per_status_cb.currentText(),population=self.population, projection=self.status_operation.currentText(), groupby_columns=self.groupby_cols)

		self.subtable = TableUI(group_table,f"Group by tracks: {self.projection_mode}", plot_mode="static")
		self.subtable.show()

		self.projectionWidget.close()

	# def groupby_track_table(self):

	# 	"""
		
	# 	Perform a time average across each track for all features

	# 	"""

	# 	self.subtable = TrajectoryTablePanel(self.data.groupby("TRACK_ID").mean(),"Group by tracks", plot_mode="scatter")
	# 	self.subtable.show()

	def _createMenuBar(self):
		menuBar = self.menuBar()
		self.fileMenu = QMenu("&File", self)
		menuBar.addMenu(self.fileMenu)
		self.editMenu = QMenu("&Edit", self)
		menuBar.addMenu(self.editMenu)
		self.mathMenu = QMenu('&Math', self)
		menuBar.addMenu(self.mathMenu)

	def save_as_csv(self):
		options = QFileDialog.Options()
		options |= QFileDialog.ReadOnly
		file_name, _ = QFileDialog.getSaveFileName(self, "Save as .csv", "","CSV Files (*.csv);;All Files (*)", options=options)
		if file_name:
			if not file_name.endswith(".csv"):
				file_name += ".csv"
			self.data.to_csv(file_name, index=False)

	def test_bool(self, array):
		if array.dtype=="bool":
			return np.array(array, dtype=int)
		else:
			return array

	def plot_instantaneous(self):
		
		if self.plot_mode=='plot_track_signals':
			self.plot_mode = 'static'
			self.plot()
			self.plot_mode = 'plot_track_signals'
		elif self.plot_mode=="static":
			self.plot()

	def plot(self):
		if self.plot_mode == "static":
	
			x = self.table_view.selectedIndexes()
			col_idx = [l.column() for l in x]
			row_idx = [l.row() for l in x]
			column_names = self.data.columns
			unique_cols = np.unique(col_idx)

			if len(unique_cols)==1 or len(unique_cols)==0:
				self.set_1D_plot_params()

			if len(unique_cols) == 2:

				print("two columns, plot mode")
				x1 = self.test_bool(self.data.iloc[row_idx, unique_cols[0]])
				x2 = self.test_bool(self.data.iloc[row_idx, unique_cols[1]])

				self.fig, self.ax = plt.subplots(1, 1, figsize=(4,3))
				self.scatter_wdw = FigureCanvas(self.fig, title="scatter")
				self.ax.clear()
				self.ax.scatter(x1,x2)
				self.ax.set_xlabel(column_names[unique_cols[0]])
				self.ax.set_ylabel(column_names[unique_cols[1]])
				plt.tight_layout()
				self.fig.set_facecolor('none')  # or 'None'
				self.fig.canvas.setStyleSheet("background-color: transparent;")
				self.scatter_wdw.canvas.draw()
				self.scatter_wdw.show()

			else:
				print("please select less columns")

		elif self.plot_mode == "plot_timeseries":
			print("mode plot frames")
			x = self.table_view.selectedIndexes()
			col_idx = np.array([l.column() for l in x])
			row_idx = np.array([l.row() for l in x])
			column_names = self.data.columns
			unique_cols = np.unique(col_idx)

			self.fig, self.ax = plt.subplots(1, 1, figsize=(4, 3))
			self.plot_wdw = FigureCanvas(self.fig, title="scatter")
			self.ax.clear()
			for k in range(len(unique_cols)):
				row_idx_i = row_idx[np.where(col_idx == unique_cols[k])[0]]
				y = self.data.iloc[row_idx_i, unique_cols[k]]
				self.ax.plot(self.data["timeline"][row_idx_i], y, label=column_names[unique_cols[k]])

			self.ax.legend()
			self.ax.set_xlabel("time [frame]")
			self.ax.set_ylabel(self.title)
			plt.tight_layout()
			self.fig.set_facecolor('none')  # or 'None'
			self.fig.canvas.setStyleSheet("background-color: transparent;")
			self.plot_wdw.canvas.draw()
			plt.show()

		elif self.plot_mode == "plot_track_signals":

			print("mode plot track signals")
			print('we plot here')

			x = self.table_view.selectedIndexes()
			col_idx = np.array([l.column() for l in x])
			row_idx = np.array([l.row() for l in x])
			column_names = self.data.columns
			unique_cols = np.unique(col_idx)

			if len(unique_cols) > 2:
				fig,ax = plt.subplots(1, 1, figsize=(7, 5.5))
				for k in range(len(unique_cols)):

					row_idx_i = row_idx[np.where(col_idx == unique_cols[k])[0]]
					y = self.data.iloc[row_idx_i, unique_cols[k]]
					print(unique_cols[k])
					for w,well_group in self.data.groupby('well_name'):
						for pos,pos_group in well_group.groupby('pos_name'):
							for tid,group_track in pos_group.groupby(self.groupby_cols[1:]):
								ax.plot(group_track["FRAME"], group_track[column_names[unique_cols[k]]],label=column_names[unique_cols[k]])
					#ax.plot(self.data["FRAME"][row_idx_i], y, label=column_names[unique_cols[k]])
				ax.legend()
				ax.set_xlabel("time [frame]")
				ax.set_ylabel(self.title)
				plt.tight_layout()
				plt.show(block=False)

			if len(unique_cols) == 2:

				self.fig, self.ax = plt.subplots(1, 1, figsize=(4, 3))
				self.scatter_wdw = FigureCanvas(self.fig, title="scatter")
				self.ax.clear()
				for tid,group in self.data.groupby(self.groupby_cols[1:]):
					self.ax.plot(group[column_names[unique_cols[0]]], group[column_names[unique_cols[1]]], marker="o")
				self.ax.set_xlabel(column_names[unique_cols[0]])
				self.ax.set_ylabel(column_names[unique_cols[1]])
				plt.tight_layout()
				self.fig.set_facecolor('none')  # or 'None'
				self.fig.canvas.setStyleSheet("background-color: transparent;")
				self.scatter_wdw.canvas.draw()
				self.scatter_wdw.show()

			if len(unique_cols) == 1:
				
				self.fig, self.ax = plt.subplots(1, 1, figsize=(4, 3))
				self.plot_wdw = FigureCanvas(self.fig, title="scatter")
				self.ax.clear()
				
				# if 't0' in list(self.data.columns):
				# 	ref_time_col = 't0'
				# else:
				# 	ref_time_col = 'FRAME'

				for w,well_group in self.data.groupby('well_name'):
					for pos,pos_group in well_group.groupby('pos_name'):
						for tid,group_track in pos_group.groupby(self.groupby_cols[1:]):
							self.ax.plot(group_track["FRAME"], group_track[column_names[unique_cols[0]]],c="k", alpha = 0.1)
				self.ax.set_xlabel(r"$t$ [frame]")
				self.ax.set_ylabel(column_names[unique_cols[0]])
				plt.tight_layout()
				self.fig.set_facecolor('none')  # or 'None'
				self.fig.canvas.setStyleSheet("background-color: transparent;")
				self.plot_wdw.canvas.draw()
				self.plot_wdw.show()


from PyQt5.QtWidgets import QMessageBox, QComboBox, \
	QCheckBox, QLineEdit, QVBoxLayout, QWidget, QLabel, QHBoxLayout, QPushButton
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QIcon, QDoubleValidator

from celldetective.gui.gui_utils import center_window
from celldetective.gui.generic_signal_plot import GenericSignalPlotWidget
from superqt import QLabeledSlider, QColormapComboBox, QSearchableComboBox
from celldetective.utils import get_software_location, _extract_labels_from_config
from celldetective.io import load_experiment_tables
from celldetective.signals import mean_signal
import numpy as np
import os
import matplotlib.pyplot as plt
plt.rcParams['svg.fonttype'] = 'none'
from glob import glob
from natsort import natsorted
import pandas as pd
import math
from celldetective.gui import Styles
from matplotlib import colormaps


class ConfigSignalPlot(QWidget, Styles):
	
	"""
	UI to set survival instructions.

	"""

	def __init__(self, parent_window=None):
		
		super().__init__()
		self.parent_window = parent_window
		self.setWindowTitle("Configure signal plot")
		self.setWindowIcon(QIcon(os.sep.join(['celldetective','icons','mexican-hat.png'])))
		self.exp_dir = self.parent_window.exp_dir
		self.soft_path = get_software_location()		
		self.exp_config = self.exp_dir +"config.ini"
		self.wells = np.array(self.parent_window.parent_window.wells,dtype=str)
		self.well_labels = _extract_labels_from_config(self.exp_config,len(self.wells))
		self.FrameToMin = self.parent_window.parent_window.FrameToMin
		self.float_validator = QDoubleValidator()
		self.target_class = [0,1]
		self.show_ci = True
		self.show_cell_lines = False
		self.ax2=None
		self.auto_close = False

		self.well_option = self.parent_window.parent_window.well_list.currentIndex()
		self.position_option = self.parent_window.parent_window.position_list.currentIndex()
		self.interpret_pos_location()

		self.screen_height = self.parent_window.parent_window.parent_window.screen_height
		center_window(self)
		self.populate_widget()

		if self.auto_close:
			self.close()
		self.setAttribute(Qt.WA_DeleteOnClose)

	def interpret_pos_location(self):
		
		"""
		Read the well/position selection from the control panel to decide which data to load
		Set position_indices to None if all positions must be taken

		"""
		
		if self.well_option==len(self.wells):
			self.well_indices = np.arange(len(self.wells))
		else:
			self.well_indices = np.array([self.well_option],dtype=int)

		if self.position_option==0:
			self.position_indices = None
		else:
			self.position_indices = np.array([self.position_option],dtype=int)

	def populate_widget(self):

		"""
		Create the multibox design.

		"""
		
		# Create button widget and layout
		main_layout = QVBoxLayout()
		self.setLayout(main_layout)
		main_layout.setContentsMargins(30,30,30,30)
		panel_title = QLabel('Options')
		panel_title.setStyleSheet("""
			font-weight: bold;
			padding: 0px;
			""")
		main_layout.addWidget(panel_title, alignment=Qt.AlignCenter)

		labels = [QLabel('population: '), QLabel('class: '), QLabel('time of\ninterest: '), QLabel('cmap: ')]
		self.cb_options = [['targets','effectors'],['class'], ['t0'], []]
		self.cbs = [QComboBox() for i in range(len(labels))]
		self.cbs[-1] = QColormapComboBox()
		
		self.cbs[0].currentIndexChanged.connect(self.set_classes_and_times)

		choice_layout = QVBoxLayout()
		choice_layout.setContentsMargins(20,20,20,20)
		for i in range(len(labels)):
			hbox = QHBoxLayout()
			hbox.addWidget(labels[i], 33)
			hbox.addWidget(self.cbs[i],66)
			if i < len(labels)-1:
				self.cbs[i].addItems(self.cb_options[i])
			choice_layout.addLayout(hbox)
		
		all_cms = list(colormaps)
		for cm in all_cms:
			try:
				self.cbs[-1].addColormap(cm)
			except:
				pass

		self.cbs[0].setCurrentIndex(1)
		self.cbs[0].setCurrentIndex(0)

		self.abs_time_checkbox = QCheckBox('absolute time')
		self.frame_slider = QLabeledSlider()
		self.frame_slider.setSingleStep(1)
		self.frame_slider.setOrientation(1)
		self.frame_slider.setRange(0,self.parent_window.parent_window.len_movie)
		self.frame_slider.setValue(0)
		self.frame_slider.setEnabled(False)
		slider_hbox = QHBoxLayout()
		slider_hbox.addWidget(self.abs_time_checkbox, 33)
		slider_hbox.addWidget(self.frame_slider, 66)
		choice_layout.addLayout(slider_hbox)
		main_layout.addLayout(choice_layout)

		self.abs_time_checkbox.stateChanged.connect(self.switch_ref_time_mode)

		select_layout = QHBoxLayout()
		select_layout.addWidget(QLabel('select cells\nwith query: '), 33)
		self.query_le = QLineEdit()
		select_layout.addWidget(self.query_le, 66)
		main_layout.addLayout(select_layout)

		time_calib_layout = QHBoxLayout()
		time_calib_layout.setContentsMargins(20,20,20,20)
		time_calib_layout.addWidget(QLabel('time calibration\n(frame to min)'), 33)
		self.time_calibration_le = QLineEdit(str(self.FrameToMin).replace('.',','))
		self.time_calibration_le.setValidator(self.float_validator)
		time_calib_layout.addWidget(self.time_calibration_le, 66)
		#time_calib_layout.addWidget(QLabel(' min'))
		main_layout.addLayout(time_calib_layout)

		self.submit_btn = QPushButton('Submit')
		self.submit_btn.setStyleSheet(self.button_style_sheet)
		self.submit_btn.clicked.connect(self.process_signal)
		main_layout.addWidget(self.submit_btn)

		#self.populate_left_panel()
		#grid.addLayout(self.left_side, 0, 0, 1, 1)

		# self.setCentralWidget(self.scroll_area)
		# self.show()

	def set_classes_and_times(self):

		# Look for all classes and times
		tables = natsorted(glob(self.exp_dir+os.sep.join(['W*','*','output','tables',f'trajectories_*.csv'])))
		self.all_columns = []
		for tab in tables:
			cols = pd.read_csv(tab, nrows=1).columns.tolist()
			self.all_columns.extend(cols)
		self.all_columns = np.unique(self.all_columns)
		class_idx = np.array([s.startswith('class_') for s in self.all_columns])
		time_idx = np.array([s.startswith('t_') for s in self.all_columns])
		
		try:
			class_columns = list(self.all_columns[class_idx])
			time_columns = list(self.all_columns[time_idx])
		except:
			print('columns not found')
			self.auto_close = True
			return None
		
		self.cbs[2].clear()
		self.cbs[2].addItems(np.unique(self.cb_options[2]+time_columns))

		self.cbs[1].clear()
		self.cbs[1].addItems(np.unique(self.cb_options[1]+class_columns))

	def ask_for_feature(self):

		cols = np.array(list(self.df.columns))
		is_number = np.vectorize(lambda x: np.issubdtype(x, np.number))
		feats = cols[is_number(self.df.dtypes)]

		self.feature_choice_widget = QWidget()
		self.feature_choice_widget.setWindowTitle("Select numeric feature")
		layout = QVBoxLayout()
		self.feature_choice_widget.setLayout(layout)
		self.feature_cb = QComboBox()
		self.feature_cb.addItems(feats)
		hbox = QHBoxLayout()
		hbox.addWidget(QLabel('feature: '), 33)
		hbox.addWidget(self.feature_cb, 66)
		layout.addLayout(hbox)

		self.set_feature_btn = QPushButton('set')
		self.set_feature_btn.clicked.connect(self.compute_signals)
		layout.addWidget(self.set_feature_btn)
		self.feature_choice_widget.show()
		center_window(self.feature_choice_widget)

	def ask_for_features(self):

		cols = np.array(list(self.df.columns))
		is_number = np.vectorize(lambda x: np.issubdtype(x, np.number))
		feats = cols[is_number(self.df.dtypes)]

		self.feature_choice_widget = QWidget()
		self.feature_choice_widget.setWindowTitle("Select numeric feature")
		layout = QVBoxLayout()
		self.feature_choice_widget.setLayout(layout)
		self.feature_cb = QSearchableComboBox()
		self.feature_cb.addItems(feats)
		hbox = QHBoxLayout()
		hbox.addWidget(QLabel('feature: '), 33)
		hbox.addWidget(self.feature_cb, 66)
		#hbox.addWidget((QLabel('Plot two features')))
		layout.addLayout(hbox)

		self.set_feature_btn = QPushButton('set')
		self.set_feature_btn.setStyleSheet(self.button_style_sheet)
		self.set_feature_btn.clicked.connect(self.compute_signals)
		layout.addWidget(self.set_feature_btn)
		self.feature_choice_widget.show()
		center_window(self.feature_choice_widget)

	# def enable_second_feature(self):
	# 	if self.checkBox_feature.isChecked():
	# 		self.feature_two_cb.setEnabled(True)
	# 	else:
	# 		self.feature_two_cb.setEnabled(False)
	
	def compute_signals(self):

		if self.df is not None:

			try:
				query_text = self.query_le.text()
				if query_text != '':
					self.df = self.df.query(query_text)
			except Exception as e:
				print(e, ' The query is misunderstood and will not be applied...')

			self.feature_selected = self.feature_cb.currentText()
			self.feature_choice_widget.close()
			self.compute_signal_functions()
			self.interpret_pos_location()
			self.plot_window = GenericSignalPlotWidget(parent_window=self, df=self.df, df_pos_info = self.df_pos_info, df_well_info = self.df_well_info, feature_selected=self.feature_selected, title='plot signals')
			self.plot_window.show()

	def process_signal(self):

		self.FrameToMin = float(self.time_calibration_le.text().replace(',','.'))
		print(f'Time calibration set to 1 frame =  {self.FrameToMin} min...')

		# read instructions from combobox options
		self.load_available_tables()
		class_col = self.cbs[1].currentText()	

		if self.df is not None:

			if class_col not in list(self.df.columns):
				msgBox = QMessageBox()
				msgBox.setIcon(QMessageBox.Warning)
				msgBox.setText("The class of interest could not be found in the data. Abort.")
				msgBox.setWindowTitle("Warning")
				msgBox.setStandardButtons(QMessageBox.Ok)
				returnValue = msgBox.exec()
				if returnValue == QMessageBox.Ok:
					return None
			else:
				self.ask_for_features()
		else:
			return None

		#self.plotvbox.addWidget(self.line_choice_widget, alignment=Qt.AlignCenter)

	def load_available_tables(self):

		"""
		Load the tables of the selected wells/positions from the control Panel for the population of interest

		"""

		self.well_option = self.parent_window.parent_window.well_list.currentIndex()
		if self.well_option==len(self.wells):
			wo = '*'
		else:
			wo = self.well_option
		self.position_option = self.parent_window.parent_window.position_list.currentIndex()
		if self.position_option==0:
			po = '*'
		else:
			po = self.position_option - 1

		self.df, self.df_pos_info = load_experiment_tables(self.exp_dir, well_option=wo, position_option=po, population=self.cbs[0].currentText(), return_pos_info=True)

		if self.df is None:
			
			print('No table could be found...')
			msgBox = QMessageBox()
			msgBox.setIcon(QMessageBox.Warning)
			msgBox.setText("No table could be found to compute survival...")
			msgBox.setWindowTitle("Warning")
			msgBox.setStandardButtons(QMessageBox.Ok)
			returnValue = msgBox.exec()
			if returnValue == QMessageBox.Ok:
				self.close()
				return None
			else:
				self.close()
				return None
		else:
			self.df_well_info = self.df_pos_info.loc[:,['well_path', 'well_index', 'well_name', 'well_number', 'well_alias']].drop_duplicates()

	def compute_signal_functions(self):

		# REPLACE EVRYTHING WITH MEAN_SIGNAL FUNCTION

		# Per position signal
		max_time = int(self.df.FRAME.max()) + 1
		class_col = self.cbs[1].currentText()
		time_col = self.cbs[2].currentText()
		if self.abs_time_checkbox.isChecked():
			time_col = self.frame_slider.value()

		for block,movie_group in self.df.groupby(['well','position']):

			well_signal_mean, well_std_mean, timeline_all, matrix_all = mean_signal(movie_group, self.feature_selected, class_col, time_col=time_col, class_value=[0,1], return_matrix=True, forced_max_duration=max_time)
			well_signal_event, well_std_event, timeline_event, matrix_event = mean_signal(movie_group, self.feature_selected, class_col, time_col=time_col, class_value=[0], return_matrix=True, forced_max_duration=max_time)		
			well_signal_no_event, well_std_no_event, timeline_no_event, matrix_no_event = mean_signal(movie_group, self.feature_selected, class_col, time_col=time_col, class_value=[1], return_matrix=True, forced_max_duration=max_time)
			self.mean_plots_timeline = timeline_all

			self.df_pos_info.loc[self.df_pos_info['pos_path'] == block[1], 'signal'] = [
				{'mean_all': well_signal_mean, 'std_all': well_std_mean, 'matrix_all': matrix_all,
				 'mean_event': well_signal_event, 'std_event': well_std_event,
				 'matrix_event': matrix_event, 'mean_no_event': well_signal_no_event,
				 'std_no_event': well_std_no_event, 'matrix_no_event': matrix_no_event,
				 'timeline': self.mean_plots_timeline}]

		# Per well
		for well,well_group in self.df.groupby('well'):

			well_signal_mean, well_std_mean, timeline_all, matrix_all = mean_signal(well_group, self.feature_selected, class_col, time_col=time_col, class_value=[0,1], return_matrix=True, forced_max_duration=max_time)
			well_signal_event, well_std_event, timeline_event, matrix_event = mean_signal(well_group, self.feature_selected, class_col, time_col=time_col, class_value=[0], return_matrix=True, forced_max_duration=max_time)			
			well_signal_no_event, well_std_no_event, timeline_no_event, matrix_no_event = mean_signal(well_group, self.feature_selected, class_col, time_col=time_col, class_value=[1], return_matrix=True, forced_max_duration=max_time)
			
			self.df_well_info.loc[self.df_well_info['well_path']==well,'signal'] = [{'mean_all': well_signal_mean, 'std_all': well_std_mean,'matrix_all': matrix_all,'mean_event': well_signal_event, 'std_event': well_std_event,
																					'matrix_event': matrix_event,'mean_no_event': well_signal_no_event, 'std_no_event': well_std_no_event, 'matrix_no_event': matrix_no_event, 'timeline':  self.mean_plots_timeline}]

		self.df_pos_info.loc[:,'select'] = True
		self.df_well_info.loc[:,'select'] = True


	def generate_synchronized_matrix(self, well_group, feature_selected, cclass, max_time):
		
		if isinstance(cclass,int):
			cclass = [cclass]

		n_cells = len(well_group.groupby(['position','TRACK_ID']))
		depth = int(2*max_time + 3)
		matrix = np.zeros((n_cells, depth))
		matrix[:,:] = np.nan
		mapping = np.arange(-max_time-1, max_time+2)
		cid=0
		for block,movie_group in well_group.groupby('position'):
			for tid,track_group in movie_group.loc[movie_group[self.cbs[1].currentText()].isin(cclass)].groupby('TRACK_ID'):
				try:
					timeline = track_group['FRAME'].to_numpy().astype(int)
					feature = track_group[feature_selected].to_numpy()
					if self.checkBox_feature.isChecked():
						second_feature=track_group[self.second_feature_selected].to_numpy()
					if self.cbs[2].currentText().startswith('t') and not self.abs_time_checkbox.isChecked():
						t0 = math.floor(track_group[self.cbs[2].currentText()].to_numpy()[0])
						timeline -= t0
					elif self.cbs[2].currentText()=='first detection' and not self.abs_time_checkbox.isChecked():

						if 'area' in list(track_group.columns):
							print('area in list')
							feat = track_group['area'].values
						else:
							feat = feature

						first_detection = timeline[feat==feat][0]
						timeline -= first_detection
						print(first_detection, timeline)

					elif self.abs_time_checkbox.isChecked():
						timeline -= int(self.frame_slider.value())

					loc_t = [np.where(mapping==t)[0][0] for t in timeline]
					matrix[cid,loc_t] = feature
					if second_feature:
						matrix[cid,loc_t+1]=second_feature
					print(timeline, loc_t)

					cid+=1
				except:
					pass
		return matrix		

	def col_mean(self, matrix):

		mean_line = np.zeros(matrix.shape[1])
		mean_line[:] = np.nan
		std_line = np.copy(mean_line)

		for k in range(matrix.shape[1]):
			values = matrix[:,k]
			#values = values[values!=0]
			if len(values[values==values])>2:
				mean_line[k] = np.nanmean(values)
				std_line[k] = np.nanstd(values)

		return mean_line, std_line


	def switch_ref_time_mode(self):
		if self.abs_time_checkbox.isChecked():
			self.frame_slider.setEnabled(True)
			self.cbs[-2].setEnabled(False)
		else:
			self.frame_slider.setEnabled(False)
			self.cbs[-2].setEnabled(True)


"""
Copright © 2022 Laboratoire Adhesion et Inflammation, Authored by Remy Torro.
"""

import argparse
import os
import json
from celldetective.io import auto_load_number_of_frames, load_frames
from celldetective.segmentation import segment_frame_from_thresholds
from celldetective.utils import _extract_channel_indices_from_config, ConfigSectionMap, _extract_nbr_channels_from_config, _get_img_num_per_channel, extract_experiment_channels
from pathlib import Path, PurePath
from glob import glob
from shutil import rmtree
from tqdm import tqdm
import numpy as np
from csbdeep.io import save_tiff_imagej_compatible
import gc
from art import tprint

tprint("Segment")

parser = argparse.ArgumentParser(description="Segment a movie in position with a threshold pipeline",
								formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument('-p',"--position", required=True, help="Path to the position")
parser.add_argument('-c',"--config", required=True,help="Threshold instructions")
parser.add_argument("--mode", default="target", choices=["target","effector","targets","effectors"],help="Cell population of interest")
parser.add_argument("--threads", default="1",help="Number of parallel threads")

args = parser.parse_args()
process_arguments = vars(args)
pos = str(process_arguments['position'])
mode = str(process_arguments['mode'])
n_threads = int(process_arguments['threads'])

threshold_instructions = str(process_arguments['config'])
equalize = False

if os.path.exists(threshold_instructions):
	with open(threshold_instructions, 'r') as f:
		threshold_instructions = json.load(f)
		required_channels = [threshold_instructions['target_channel']]
		if 'equalize_reference' in threshold_instructions:
			equalize_info = threshold_instructions['equalize_reference']
			equalize = equalize_info[0]
			equalize_time = equalize_info[1]

else:
	print('The configuration path is not valid. Abort.')
	os.abort()

print('The following instructions were successfully loaded: ', threshold_instructions)

if mode.lower()=="target" or mode.lower()=="targets":
	label_folder = "labels_targets"
elif mode.lower()=="effector" or mode.lower()=="effectors":
	label_folder = "labels_effectors"

# Locate experiment config
parent1 = Path(pos).parent
expfolder = parent1.parent
config = PurePath(expfolder,Path("config.ini"))
assert os.path.exists(config),'The configuration file for the experiment could not be located. Abort.'
print("Configuration file: ",config)


channel_indices = _extract_channel_indices_from_config(config, required_channels)
# need to abort if channel not found
print(f'Required channels: {required_channels} located at channel indices {channel_indices}.')

threshold_instructions.update({'target_channel': channel_indices[0]})

movie_prefix = ConfigSectionMap(config,"MovieSettings")["movie_prefix"]
len_movie = float(ConfigSectionMap(config,"MovieSettings")["len_movie"])
channel_names, channel_indices = extract_experiment_channels(config)
threshold_instructions.update({'channel_names': channel_names})

# Try to find the file
try:
	file = glob(pos+f"movie/{movie_prefix}*.tif")[0]
except IndexError:
	print('Movie could not be found. Check the prefix.')
	os.abort()

len_movie_auto = auto_load_number_of_frames(file)
if len_movie_auto is not None:
	len_movie = len_movie_auto

nbr_channels = _extract_nbr_channels_from_config(config)
print(f'Number of channels in the input movie: {nbr_channels}')
img_num_channels = _get_img_num_per_channel(np.arange(nbr_channels), len_movie, nbr_channels)

# If everything OK, prepare output, load models
print('Erasing previous segmentation folder.')
if os.path.exists(os.sep.join([pos,label_folder])):
	rmtree(os.sep.join([pos,label_folder]))
os.mkdir(os.sep.join([pos,label_folder]))
print(f'Folder {os.sep.join([pos,label_folder])} successfully generated.')

if equalize:
	f_reference = load_frames(img_num_channels[:,equalize_time], file, scale=None, normalize_input=False)
	f_reference = f_reference[:,:,threshold_instructions['target_channel']]
else:
	f_reference = None

threshold_instructions.update({'equalize_reference': f_reference})
print(threshold_instructions)

# Loop over all frames and segment
def segment_index(indices):

	for t in tqdm(indices,desc="frame"):
		
		# Load channels at time t
		f = load_frames(img_num_channels[:,t], file, scale=None, normalize_input=False)
		mask = segment_frame_from_thresholds(f, **threshold_instructions)
		save_tiff_imagej_compatible(os.sep.join([pos, label_folder, f"{str(t).zfill(4)}.tif"]), mask.astype(np.uint16), axes='YX')

		del f;
		del mask;
		gc.collect()

import concurrent.futures

# Multithreading
indices = list(range(img_num_channels.shape[1]))
chunks = np.array_split(indices, n_threads)

with concurrent.futures.ThreadPoolExecutor() as executor:
	executor.map(segment_index, chunks)

# indices = list(range(img_num_channels.shape[1]))
# chunks = np.array_split(indices, n_threads)
# threads = []
# for i in range(n_threads):
# 	thread_i = threading.Thread(target=segment_index, args=[chunks[i]])
# 	threads.append(thread_i)
# for th in threads:
# 	th.start()
# for th in threads:
# 	th.join()

print('Done.')
gc.collect()







import unittest
import matplotlib.pyplot as plt
import numpy as np
import os

# class TestPatchMask(unittest.TestCase):

# 	@classmethod
# 	def setUpClass(self):
# 		self.radius = 3

# 	def test_correct_shape(self):
# 		self.patch = create_patch_mask(self.radius, self.radius)
# 		self.assertEqual(self.patch.shape,(3,3))

# 	def test_correct_ring(self):
# 		self.patch = create_patch_mask(5, 5,radius=[1,2])
# 		self.assertFalse(self.patch[2,2])

# class TestRemoveRedundantFeatures(unittest.TestCase):

# 	@classmethod
# 	def setUpClass(self):
# 		self.list_a = ['feat1','feat2','feat3','feat4','intensity_mean']
# 		self.list_b = ['feat5','feat2','feat1','feat6','test_channel_mean']
# 		self.expected = ['feat3','feat4']

# 	def test_remove_red_features(self):
# 		self.assertEqual(remove_redundant_features(self.list_a, self.list_b, channel_names=['test_channel']), self.expected)


# class TestExtractChannelIndices(unittest.TestCase):

# 	@classmethod
# 	def setUpClass(self):
# 		self.channels = ['ch1','ch2','ch3','ch4']
# 		self.required_channels = ['ch4','ch2']
# 		self.expected_indices = [3,1]

# 	def test_extracted_channels_are_correct(self):
# 		self.assertEqual(list(_extract_channel_indices(self.channels, self.required_channels)), self.expected_indices)


# class TestImgIndexPerChannel(unittest.TestCase):

# 	@classmethod
# 	def setUpClass(self):
# 		self.channels_indices = [1]
# 		self.len_movie = 5
# 		self.nbr_channels = 3
# 		self.expected_indices = [1,4,7,10,13]

# 	def test_index_sequence_is_correct(self):
# 		self.assertEqual(list(_get_img_num_per_channel(self.channels_indices, self.len_movie, self.nbr_channels)[0]), self.expected_indices)


# class TestSplitArrayByRatio(unittest.TestCase):

# 	@classmethod
# 	def setUpClass(self):
# 		self.array_length = 100
# 		self.array = np.ones(self.array_length)

# 	def test_ratio_split_is_correct(self):
# 		split_array = split_by_ratio(self.array,0.5,0.25,0.1)
# 		self.assertTrue(np.all([len(split_array[0])==50, len(split_array[1])==25, len(split_array[2])==10]))


# if __name__=="__main__":
# 	unittest.main()


from PyQt5.QtWidgets import QMessageBox, QComboBox, QLineEdit, QVBoxLayout, QWidget, QLabel, QHBoxLayout, QPushButton
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QDoubleValidator
from celldetective.gui.gui_utils import center_window
from superqt import QColormapComboBox
from celldetective.gui.generic_signal_plot import SurvivalPlotWidget
from celldetective.utils import get_software_location, _extract_labels_from_config
from celldetective.io import load_experiment_tables
import numpy as np
import os
import matplotlib.pyplot as plt
plt.rcParams['svg.fonttype'] = 'none'
from glob import glob
import pandas as pd
from celldetective.gui import Styles
from matplotlib import colormaps
from celldetective.events import compute_survival

class ConfigSurvival(QWidget, Styles):
	
	"""
	UI to set survival instructions.

	"""

	def __init__(self, parent_window=None):
		
		super().__init__()
		self.parent_window = parent_window
		self.setWindowTitle("Configure survival")
		self.setWindowIcon(self.celldetective_icon)

		self.exp_dir = self.parent_window.exp_dir
		self.soft_path = get_software_location()		
		self.exp_config = self.exp_dir +"config.ini"
		self.wells = np.array(self.parent_window.parent_window.wells,dtype=str)
		self.well_labels = _extract_labels_from_config(self.exp_config,len(self.wells))
		self.FrameToMin = self.parent_window.parent_window.FrameToMin
		self.float_validator = QDoubleValidator()
		self.auto_close = False

		self.well_option = self.parent_window.parent_window.well_list.currentIndex()
		self.position_option = self.parent_window.parent_window.position_list.currentIndex()
		self.interpret_pos_location()
		#self.config_path = self.exp_dir + self.config_name

		self.screen_height = self.parent_window.parent_window.parent_window.screen_height
		center_window(self)

		self.setMinimumWidth(350)
		self.populate_widget()
		if self.auto_close:
			self.close()
		
		self.setAttribute(Qt.WA_DeleteOnClose)

	def interpret_pos_location(self):
		
		"""
		Read the well/position selection from the control panel to decide which data to load
		Set position_indices to None if all positions must be taken

		"""
		
		if self.well_option==len(self.wells):
			self.well_indices = np.arange(len(self.wells))
		else:
			self.well_indices = np.array([self.well_option],dtype=int)

		if self.position_option==0:
			self.position_indices = None
		else:
			self.position_indices = np.array([self.position_option],dtype=int)


	def populate_widget(self):

		"""
		Create the multibox design.

		"""
		
		# Create button widget and layout
		main_layout = QVBoxLayout()
		self.setLayout(main_layout)
		main_layout.setContentsMargins(30,30,30,30)

		panel_title = QLabel('Options')
		panel_title.setStyleSheet("""
			font-weight: bold;
			padding: 0px;
			""")
		main_layout.addWidget(panel_title, alignment=Qt.AlignCenter)


		labels = [QLabel('population: '), QLabel('time of\nreference: '), QLabel('time of\ninterest: '), QLabel('cmap: ')] #QLabel('class: '), 
		self.cb_options = [['targets','effectors'], ['0'], [], []] #['class'], 
		self.cbs = [QComboBox() for i in range(len(labels))]

		self.cbs[-1] = QColormapComboBox()
		self.cbs[0].currentIndexChanged.connect(self.set_classes_and_times)

		choice_layout = QVBoxLayout()
		choice_layout.setContentsMargins(20,20,20,20)
		for i in range(len(labels)):
			hbox = QHBoxLayout()
			hbox.addWidget(labels[i], 33)
			hbox.addWidget(self.cbs[i],66)
			if i < len(labels)-1:
				self.cbs[i].addItems(self.cb_options[i])
			choice_layout.addLayout(hbox)

		all_cms = list(colormaps)
		for cm in all_cms:
			try:
				self.cbs[-1].addColormap(cm)
			except:
				pass

		main_layout.addLayout(choice_layout)

		select_layout = QHBoxLayout()
		select_layout.addWidget(QLabel('select cells\nwith query: '), 33)
		self.query_le = QLineEdit()
		select_layout.addWidget(self.query_le, 66)
		main_layout.addLayout(select_layout)

		self.cbs[0].setCurrentIndex(0)
		self.cbs[1].setCurrentText('t_firstdetection')

		time_calib_layout = QHBoxLayout()
		time_calib_layout.setContentsMargins(20,20,20,20)
		time_calib_layout.addWidget(QLabel('time calibration\n(frame to min)'), 33)
		self.time_calibration_le = QLineEdit(str(self.FrameToMin).replace('.',','))
		self.time_calibration_le.setValidator(self.float_validator)
		time_calib_layout.addWidget(self.time_calibration_le, 66)
		#time_calib_layout.addWidget(QLabel(' min'))
		main_layout.addLayout(time_calib_layout)

		self.submit_btn = QPushButton('Submit')
		self.submit_btn.setStyleSheet(self.button_style_sheet)
		self.submit_btn.clicked.connect(self.process_survival)
		main_layout.addWidget(self.submit_btn)

		#self.populate_left_panel()
		#grid.addLayout(self.left_side, 0, 0, 1, 1)

		# self.setCentralWidget(self.scroll_area)
		# self.show()

	def set_classes_and_times(self):

		# Look for all classes and times
		tables = glob(self.exp_dir+os.sep.join(['W*','*','output','tables',f'trajectories_*.csv']))
		self.all_columns = []
		for tab in tables:
			cols = pd.read_csv(tab, nrows=1,encoding_errors='ignore').columns.tolist()
			self.all_columns.extend(cols)

		self.all_columns = np.unique(self.all_columns)
		#class_idx = np.array([s.startswith('class_') for s in self.all_columns])
		time_idx = np.array([s.startswith('t_') for s in self.all_columns])

		try:
			time_columns = list(self.all_columns[time_idx])
		except:
			print('no column starts with t')
			self.auto_close = True
			return None

		self.cbs[1].clear()
		self.cbs[1].addItems(np.unique(self.cb_options[1]+time_columns))
		self.cbs[1].setCurrentText('t_firstdetection')

		self.cbs[2].clear()
		self.cbs[2].addItems(np.unique(self.cb_options[2]+time_columns))


	def process_survival(self):

		print('you clicked!!')
		self.FrameToMin = float(self.time_calibration_le.text().replace(',','.'))
		print(self.FrameToMin, 'set')

		self.time_of_interest = self.cbs[2].currentText()
		if self.time_of_interest=="t0":
			self.class_of_interest = "class"
		else:
			self.class_of_interest = self.time_of_interest.replace('t_','class_')

		# read instructions from combobox options
		self.load_available_tables_local()

		if self.df is not None:
			
			try:
				query_text = self.query_le.text()
				if query_text != '':
					self.df = self.df.query(query_text)
			except Exception as e:
				print(e, ' The query is misunderstood and will not be applied...')
			
			self.interpret_pos_location()
			if self.class_of_interest in list(self.df.columns) and self.cbs[2].currentText() in list(self.df.columns):
				self.compute_survival_functions()
			else:
				msgBox = QMessageBox()
				msgBox.setIcon(QMessageBox.Warning)
				msgBox.setText("The class and/or event time of interest is not found in the dataframe...")
				msgBox.setWindowTitle("Warning")
				msgBox.setStandardButtons(QMessageBox.Ok)
				returnValue = msgBox.exec()
				if returnValue == QMessageBox.Ok:
					return None
			if 'survival_fit' in list(self.df_pos_info.columns):
				self.plot_window = SurvivalPlotWidget(parent_window=self, df=self.df, df_pos_info = self.df_pos_info, df_well_info = self.df_well_info, title='plot survivals')
				self.plot_window.show()
			else:
				msgBox = QMessageBox()
				msgBox.setIcon(QMessageBox.Warning)
				msgBox.setText("No survival function was successfully computed...\nCheck your parameter choice.")
				msgBox.setWindowTitle("Warning")
				msgBox.setStandardButtons(QMessageBox.Ok)
				returnValue = msgBox.exec()
				if returnValue == QMessageBox.Ok:
					return None	

	def load_available_tables_local(self):

		"""
		Load the tables of the selected wells/positions from the control Panel for the population of interest

		"""

		self.well_option = self.parent_window.parent_window.well_list.currentIndex()
		if self.well_option==len(self.wells):
			wo = '*'
		else:
			wo = self.well_option
		self.position_option = self.parent_window.parent_window.position_list.currentIndex()
		if self.position_option==0:
			po = '*'
		else:
			po = self.position_option - 1

		self.df, self.df_pos_info = load_experiment_tables(self.exp_dir, well_option=wo, position_option=po, population=self.cbs[0].currentText(), return_pos_info=True)
		if self.df is None:
			msgBox = QMessageBox()
			msgBox.setIcon(QMessageBox.Warning)
			msgBox.setText("No table could be found.. Abort.")
			msgBox.setWindowTitle("Warning")
			msgBox.setStandardButtons(QMessageBox.Ok)
			returnValue = msgBox.exec()
			if returnValue == QMessageBox.Ok:
				return None		
		else:
			self.df_well_info = self.df_pos_info.loc[:,['well_path', 'well_index', 'well_name', 'well_number', 'well_alias']].drop_duplicates()
			#print(f"{self.df_well_info=}")

	def compute_survival_functions(self):

		# Per position survival
		for block,movie_group in self.df.groupby(['well','position']):

			ks = compute_survival(movie_group, self.class_of_interest, self.cbs[2].currentText(), t_reference=self.cbs[1].currentText(), FrameToMin=self.FrameToMin)
			if ks is not None:
				self.df_pos_info.loc[self.df_pos_info['pos_path']==block[1],'survival_fit'] = ks

		# Per well survival
		for well,well_group in self.df.groupby('well'):

			ks = compute_survival(well_group, self.class_of_interest, self.cbs[2].currentText(), t_reference=self.cbs[1].currentText(), FrameToMin=self.FrameToMin)
			if ks is not None:
				self.df_well_info.loc[self.df_well_info['well_path']==well,'survival_fit'] = ks

		self.df_pos_info.loc[:,'select'] = True
		self.df_well_info.loc[:,'select'] = True


"""
Copright © 2022 Laboratoire Adhesion et Inflammation, Authored by Remy Torro.
"""

import argparse
import datetime
import os
import json
from stardist.models import StarDist2D
from cellpose.models import CellposeModel
from celldetective.io import locate_segmentation_model, auto_load_number_of_frames, load_frames
from celldetective.utils import interpolate_nan, _estimate_scale_factor, _extract_channel_indices_from_config, ConfigSectionMap, _extract_nbr_channels_from_config, _get_img_num_per_channel
from pathlib import Path, PurePath
from glob import glob
from shutil import rmtree
from tqdm import tqdm
import numpy as np
from skimage.transform import resize
from csbdeep.io import save_tiff_imagej_compatible
import gc
from art import tprint
from scipy.ndimage import zoom


tprint("Segment")

parser = argparse.ArgumentParser(description="Segment a movie in position with the selected model",
								formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument('-p',"--position", required=True, help="Path to the position")
parser.add_argument('-m',"--model", required=True,help="Model name")
parser.add_argument("--mode", default="target", choices=["target","effector","targets","effectors"],help="Cell population of interest")
parser.add_argument("--use_gpu", default="True", choices=["True","False"],help="use GPU")
parser.add_argument("--threads", default="1",help="Number of parallel threads")

args = parser.parse_args()
process_arguments = vars(args)
pos = str(process_arguments['position'])
mode = str(process_arguments['mode'])
use_gpu = process_arguments['use_gpu']
n_threads = int(process_arguments['threads'])

if use_gpu=='True' or use_gpu=='true' or use_gpu=='1':
	use_gpu = True
	n_threads = 1  # avoid misbehavior on GPU with multithreading
else:
	use_gpu = False

if not use_gpu:
	os.environ['CUDA_VISIBLE_DEVICES'] = '-1'
n_threads = int(process_arguments['threads'])

modelname = str(process_arguments['model'])

if mode.lower()=="target" or mode.lower()=="targets":
	label_folder = "labels_targets"
elif mode.lower()=="effector" or mode.lower()=="effectors":
	label_folder = "labels_effectors"

# Locate experiment config
parent1 = Path(pos).parent
expfolder = parent1.parent
config = PurePath(expfolder,Path("config.ini"))
assert os.path.exists(config),'The configuration file for the experiment could not be located. Abort.'
print("Configuration file: ",config)

####################################
# Check model requirements #########
####################################

modelpath = os.sep.join([os.path.split(os.path.dirname(os.path.realpath(__file__)))[0],"models"])
print(modelpath)
model_complete_path = locate_segmentation_model(modelname)
if model_complete_path is None:
	print('Model could not be found. Abort.')
	os.abort()
else:
	print(f'Model successfully located in {model_complete_path}')

# load config
assert os.path.exists(model_complete_path+"config_input.json"),'The configuration for the inputs to the model could not be located. Abort.'
with open(model_complete_path+"config_input.json") as config_file:
	input_config = json.load(config_file)

# Parse target channels
required_channels = input_config["channels"]

channel_indices = _extract_channel_indices_from_config(config, required_channels)
print(f'Required channels: {required_channels} located at channel indices {channel_indices}.')
required_spatial_calibration = input_config['spatial_calibration']
print(f'Expected spatial calibration is {required_spatial_calibration}.')

normalization_percentile = input_config['normalization_percentile']
normalization_clip = input_config['normalization_clip']
normalization_values = input_config['normalization_values']

model_type = input_config['model_type']

movie_prefix = ConfigSectionMap(config,"MovieSettings")["movie_prefix"]
spatial_calibration = float(ConfigSectionMap(config,"MovieSettings")["pxtoum"])
len_movie = float(ConfigSectionMap(config,"MovieSettings")["len_movie"])

# Try to find the file
try:
	file = glob(pos+f"movie/{movie_prefix}*.tif")[0]
except IndexError:
	print('Movie could not be found. Check the prefix.')
	os.abort()

len_movie_auto = auto_load_number_of_frames(file)
if len_movie_auto is not None:
	len_movie = len_movie_auto

if model_type=='cellpose':
	diameter = input_config['diameter']
	# if diameter!=30:
	# 	required_spatial_calibration = None 	# ignore spatial calibration and use diameter
	cellprob_threshold = input_config['cellprob_threshold']
	flow_threshold = input_config['flow_threshold']

scale = _estimate_scale_factor(spatial_calibration, required_spatial_calibration)
print(f"Scale = {scale}...")

nbr_channels = _extract_nbr_channels_from_config(config)
print(f'Number of channels in the input movie: {nbr_channels}')
img_num_channels = _get_img_num_per_channel(channel_indices, int(len_movie), nbr_channels)

# If everything OK, prepare output, load models
print('Erasing previous segmentation folder.')
if os.path.exists(pos+label_folder):
	rmtree(pos+label_folder)
os.mkdir(pos+label_folder)
print(f'Folder {pos+label_folder} successfully generated.')
log=f'segmentation model: {modelname}\n'
with open(pos+f'log_{mode}.json', 'a') as f:
	f.write(f'{datetime.datetime.now()} SEGMENT \n')
	f.write(log)


# Loop over all frames and segment
def segment_index(indices):
	global scale

	if model_type=='stardist':
		model = StarDist2D(None, name=modelname, basedir=Path(model_complete_path).parent)
		model.config.use_gpu = use_gpu
		model.use_gpu = use_gpu
		print(f"StarDist model {modelname} successfully loaded.")
		scale_model = scale

	elif model_type=='cellpose':

		import torch
		if not use_gpu:
			device = torch.device("cpu")
		else:
			device = torch.device("cuda")

		model = CellposeModel(gpu=use_gpu, device=device, pretrained_model=model_complete_path+modelname, model_type=None, nchan=len(required_channels)) #diam_mean=30.0,
		if scale is None:
			scale_model = model.diam_mean / model.diam_labels
		else:
			scale_model = scale * model.diam_mean / model.diam_labels
		print(f"Diam mean: {model.diam_mean}; Diam labels: {model.diam_labels}; Final rescaling: {scale_model}...")
		print(f'Cellpose model {modelname} successfully loaded.')

	for t in tqdm(indices,desc="frame"):
		
		# Load channels at time t
		values = []
		percentiles = []
		for k in range(len(normalization_percentile)):
			if normalization_percentile[k]:
				percentiles.append(normalization_values[k])
				values.append(None)
			else:
				percentiles.append(None)
				values.append(normalization_values[k])

		f = load_frames(img_num_channels[:,t], file, scale=scale_model, normalize_input=True, normalize_kwargs={"percentiles": percentiles, 'values': values, 'clip': normalization_clip})
		f = np.moveaxis([interpolate_nan(f[:,:,c].copy()) for c in range(f.shape[-1])],0,-1)

		if np.any(img_num_channels[:,t]==-1):
			f[:,:,np.where(img_num_channels[:,t]==-1)[0]] = 0.
		

		if model_type=="stardist":
			Y_pred, details = model.predict_instances(f, n_tiles=model._guess_n_tiles(f), show_tile_progress=False, verbose=False)
			Y_pred = Y_pred.astype(np.uint16)

		elif model_type=="cellpose":

			img = np.moveaxis(f, -1, 0)
			Y_pred, _, _ = model.eval(img, diameter = diameter, cellprob_threshold=cellprob_threshold, flow_threshold=flow_threshold, channels=None, normalize=False)
			Y_pred = Y_pred.astype(np.uint16)

		if scale is not None:
			Y_pred = zoom(Y_pred, [1./scale_model,1./scale_model],order=0)

		template = load_frames(0,file,scale=1,normalize_input=False)
		if Y_pred.shape != template.shape[:2]:
			Y_pred = resize(Y_pred, template.shape[:2], order=0)

		save_tiff_imagej_compatible(pos+os.sep.join([label_folder,f"{str(t).zfill(4)}.tif"]), Y_pred, axes='YX')

		del f;
		del template;
		del Y_pred;
		gc.collect()


import concurrent.futures

# Multithreading
indices = list(range(img_num_channels.shape[1]))
chunks = np.array_split(indices, n_threads)

with concurrent.futures.ThreadPoolExecutor() as executor:
	executor.map(segment_index, chunks)

# threads = []
# for i in range(n_threads):
# 	thread_i = threading.Thread(target=segment_index, args=[chunks[i]])
# 	threads.append(thread_i)
# for th in threads:
# 	th.start()
# for th in threads:
# 	th.join()

print('Done.')

try:
	del model
except:
	pass

gc.collect()








"""
Copright © 2022 Laboratoire Adhesion et Inflammation, Authored by Remy Torro.
"""

import argparse
import datetime
import os
import json
from celldetective.io import auto_load_number_of_frames, load_frames, interpret_tracking_configuration
from celldetective.utils import extract_experiment_channels, ConfigSectionMap, _get_img_num_per_channel, extract_experiment_channels
from celldetective.measure import drop_tonal_features, measure_features
from celldetective.tracking import track
from pathlib import Path, PurePath
from glob import glob
from tqdm import tqdm
import numpy as np
import pandas as pd
import gc
import os
from natsort import natsorted
from art import tprint
from tifffile import imread

tprint("Track")

parser = argparse.ArgumentParser(description="Segment a movie in position with the selected model",
								formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument('-p',"--position", required=True, help="Path to the position")
parser.add_argument("--mode", default="target", choices=["target","effector","targets","effectors"],help="Cell population of interest")
parser.add_argument("--threads", default="1",help="Number of parallel threads")

args = parser.parse_args()
process_arguments = vars(args)
pos = str(process_arguments['position'])
mode = str(process_arguments['mode'])
n_threads = int(process_arguments['threads'])

if not os.path.exists(pos+"output"):
	os.mkdir(pos+"output")

if not os.path.exists(pos+os.sep.join(["output","tables"])):
	os.mkdir(pos+os.sep.join(["output","tables"]))

if mode.lower()=="target" or mode.lower()=="targets":
	label_folder = "labels_targets"
	instruction_file = os.sep.join(["configs", "tracking_instructions_targets.json"])
	napari_name = "napari_target_trajectories.npy"
	table_name = "trajectories_targets.csv"

elif mode.lower()=="effector" or mode.lower()=="effectors":
	label_folder = "labels_effectors"
	instruction_file = os.sep.join(["configs","tracking_instructions_effectors.json"])
	napari_name = "napari_effector_trajectories.npy"
	table_name = "trajectories_effectors.csv"

# Locate experiment config
parent1 = Path(pos).parent
expfolder = parent1.parent
config = PurePath(expfolder,Path("config.ini"))
assert os.path.exists(config),'The configuration file for the experiment could not be located. Abort.'

# from exp config fetch spatial calib, channel names
movie_prefix = ConfigSectionMap(config,"MovieSettings")["movie_prefix"]
spatial_calibration = float(ConfigSectionMap(config,"MovieSettings")["pxtoum"])
time_calibration = float(ConfigSectionMap(config,"MovieSettings")["frametomin"])
len_movie = float(ConfigSectionMap(config,"MovieSettings")["len_movie"])
shape_x = int(ConfigSectionMap(config,"MovieSettings")["shape_x"])
shape_y = int(ConfigSectionMap(config,"MovieSettings")["shape_y"])

channel_names, channel_indices = extract_experiment_channels(config)
nbr_channels = len(channel_names)

# from tracking instructions, fetch btrack config, features, haralick, clean_traj, idea: fetch custom timeline?
instr_path = PurePath(expfolder,Path(f"{instruction_file}"))
if os.path.exists(instr_path):
	print(f"Tracking instructions for the {mode} population have been successfully loaded...")
	with open(instr_path, 'r') as f:
		instructions = json.load(f)
	btrack_config = interpret_tracking_configuration(instructions['btrack_config_path'])

	if 'features' in instructions:
		features = instructions['features']
	else:
		features = None
	
	if 'mask_channels' in instructions:
		mask_channels = instructions['mask_channels']
	else:
		mask_channels = None
	
	if 'haralick_options' in instructions:
		haralick_options = instructions['haralick_options']
	else:
		haralick_options = None

	if 'post_processing_options' in instructions:
		post_processing_options = instructions['post_processing_options']
	else:
		post_processing_options = None
else:
	print('Tracking instructions could not be located... Using a standard bTrack motion model instead...')
	btrack_config = interpret_tracking_configuration(None)
	features = None
	mask_channels = None
	haralick_options = None
	post_processing_options = None

if features is None:
	features = []

# from pos fetch labels
label_path = natsorted(glob(pos+f"{label_folder}"+os.sep+"*.tif"))
if len(label_path)>0:
	print(f"Found {len(label_path)} segmented frames...")
else:
	print(f"No segmented frames have been found. Please run segmentation first. Abort...")
	os.abort()

# Do this if features or Haralick is not None, else don't need stack
try:
	file = glob(pos+os.sep.join(["movie", f"{movie_prefix}*.tif"]))[0]
except IndexError:
	print('Movie could not be found. Check the prefix. If you intended to measure texture or tone, this will not be performed.')
	file = None
	haralick_option = None
	features = drop_tonal_features(features)

len_movie_auto = auto_load_number_of_frames(file)
if len_movie_auto is not None:
	len_movie = len_movie_auto

img_num_channels = _get_img_num_per_channel(channel_indices, len_movie, nbr_channels)

#######################################
# Loop over all frames and find objects
#######################################

timestep_dataframes = []
features_log=f'features: {features}'
mask_channels_log=f'mask_channels: {mask_channels}'
haralick_option_log=f'haralick_options: {haralick_options}'
post_processing_option_log=f'post_processing_options: {post_processing_options}'
log_list=[features_log, mask_channels_log, haralick_option_log, post_processing_option_log]
log='\n'.join(log_list)

with open(pos+f'log_{mode}.json', 'a') as f:
	f.write(f'{datetime.datetime.now()} TRACK \n')
	f.write(log+"\n")

def measure_index(indices):
	for t in tqdm(indices,desc="frame"):
		
		# Load channels at time t
		img = load_frames(img_num_channels[:,t], file, scale=None, normalize_input=False)
		lbl = imread(label_path[t])

		df_props = measure_features(img, lbl, features = features+['centroid'], border_dist=None, 
										channels=channel_names, haralick_options=haralick_options, verbose=False, 
									)
		df_props.rename(columns={'centroid-1': 'x', 'centroid-0': 'y'},inplace=True)
		df_props['t'] = int(t)
		timestep_dataframes.append(df_props)

# Multithreading
indices = list(range(img_num_channels.shape[1]))
chunks = np.array_split(indices, n_threads)

import concurrent.futures

with concurrent.futures.ThreadPoolExecutor() as executor:
	executor.map(measure_index, chunks)

# threads = []
# for i in range(n_threads):
# 	thread_i = threading.Thread(target=measure_index, args=[chunks[i]])
# 	threads.append(thread_i)
# for th in threads:
# 	th.start()
# for th in threads:
# 	th.join()

df = pd.concat(timestep_dataframes)	
df.reset_index(inplace=True, drop=True)

if mask_channels is not None:
	cols_to_drop = []
	for mc in mask_channels:
		columns = df.columns
		col_contains = [mc in c for c in columns]
		to_remove = np.array(columns)[np.array(col_contains)]
		cols_to_drop.extend(to_remove)
	if len(cols_to_drop)>0:
		df = df.drop(cols_to_drop, axis=1)

# do tracking
trajectories, napari_data = track(None,
					configuration=btrack_config,
					objects=df, 
					spatial_calibration=spatial_calibration, 
					channel_names=channel_names,
					return_napari_data=True,
		  			optimizer_options = {'tm_lim': int(12e4)}, 
		  			track_kwargs={'step_size': 100}, 
		  			clean_trajectories_kwargs=post_processing_options, 
		  			volume=(shape_x, shape_y),
		  			)

# out trajectory table, create POSITION_X_um, POSITION_Y_um, TIME_min (new ones)
# Save napari data
np.save(pos+os.sep.join(['output', 'tables', napari_name]), napari_data, allow_pickle=True)
print(f"napari data successfully saved in {pos+os.sep.join(['output', 'tables'])}")

trajectories.to_csv(pos+os.sep.join(['output', 'tables', table_name]), index=False)
print(f"Table {table_name} successfully saved in {os.sep.join(['output', 'tables'])}")

if os.path.exists(pos+os.sep.join(['output', 'tables', table_name.replace('.csv','.pkl')])):
	os.remove(pos+os.sep.join(['output', 'tables', table_name.replace('.csv','.pkl')]))

del trajectories; del napari_data;
gc.collect()